

from ..imath import main


def test_imath_works(capsys):
    """"Test that imath displays its help msg"""

    try:
        main(['--help'])
    except SystemExit:
        pass

    out, err = capsys.readouterr()
    # out = out.split("\n")
    # out.sort()
    assert isinstance(out, str)



#
# Copyright 2015 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Unit test for treedict"""

from ..treedict import TreeDict


def test_treedict():
    a = TreeDict()
    a['instrument.name'] = 'iname'
    assert a['instrument.name'] == 'iname'

    de = TreeDict()
    de['val1'] = 'cal1'
    de['val2'] = 2394

    a['instrument.detector'] = de
    assert a['instrument']['detector']['val2'] == 2394
    assert a['instrument.detector.val2'] == 2394



#
# Copyright 2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Compare ADR corrections in a 3D FITS file"""
import argparse
from astropy.io import fits
from astropy.wcs import WCS
import matplotlib.pyplot as plt
import numpy as np
import sys

from .ctext import ctext

def plot_reference_wavelengths(ax, refewave1, refewave2, extname1, extname2):
    """Auxiliary function to plot reference wavelengths

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        Instance of matplotlib.axes.Axes to plot on.
    refewave1 : float
        Reference wavelength in m, corresponding to 'extname1'
    refewave2 : float
        Reference wavelength in m, corresponding to 'extname2'
    extname1 : str
        Name of the first extension.
    extname2 : str
        Name of the second extension.
    """
    if refewave1 is not None:
        ax.axvline(refewave1, color='C0', linestyle=':',
                   label=r"$\lambda_{\rm ref1}$" + f" {extname1}")
    if refewave2 is not None:
        ax.axvline(refewave2, color='C1', linestyle=':',
                   label=r"$\lambda_{\rm ref2}$" + f" {extname2}")
    if refewave1 is not None:
        ax.text(0.0, 1.05, r"$\lambda_{\rm ref1}=$" + f"{refewave1}",
                ha='left', va='bottom', color='C0', transform=ax.transAxes)
    if refewave2 is not None:
        ax.text(1.0, 1.05, r"$\lambda_{\rm ref2}=$" + f"{refewave2}",
                ha='right', va='bottom', color='C1', transform=ax.transAxes)


def compare_adr_extensions_in_3d_cube(filename, extname1, extname2=None, suptitle=None):
    """Compare 2 ADR corrections stored in 2 extensions

    Parameters
    ----------
    filename : str, file-like or `pathlib.Path`
        FITS filename to be updated.
    extname1 : str
        First extension with ADR correction data.
    extname2 : str or None
        Second extension with ADR correction data. If not `None`,
        the differences are also displayed.
    suptitle : str or None
        Centered super title to the figure. If `None`, the 'filename'
        is employed.
    """
    refewave1 = None
    refewave2 = None
    with fits.open(filename) as hdul:
        header = hdul[0].header
        if extname1 in hdul:
            table1_adrcross = hdul[extname1].data
            refewave1 = hdul[extname1].header['REFEWAVE']
        else:
            raise ValueError(f"Extension '{extname1}' not found in FITS file '{filename}'")
        if extname2 is not None:
            if extname2 in hdul:
                table2_adrcross = hdul[extname2].data
                refewave2 = hdul[extname1].header['REFEWAVE']
            else:
                raise ValueError(f"Extension '{extname2}' not found in FITS file '{filename}'")
        else:
            table2_adrcross = None

    wcs3d = WCS(header)
    naxis1, naxis2, naxis3 = wcs3d.pixel_shape
    wave = wcs3d.spectral.pixel_to_world(np.arange(naxis3))

    delta_x1 = table1_adrcross['Delta_x']
    delta_y1 = table1_adrcross['Delta_y']
    if table2_adrcross is not None:
        delta_x2 = table2_adrcross['Delta_x']
        delta_y2 = table2_adrcross['Delta_y']
        num_figures = 4
        fig, axarr = plt.subplots(nrows=2, ncols=2, figsize=(10, 7))
    else:
        delta_x2 = None
        delta_y2 = None
        num_figures = 2
        fig, axarr = plt.subplots(nrows=1, ncols=2, figsize=(10, 4.5))

    axarr = axarr.flatten()
    for iplot in range(num_figures):
        ax = axarr[iplot]
        if iplot == 0:
            ax.plot(wave, delta_x1, 'C0.', label=extname1)
            if table2_adrcross is not None:
                ax.plot(wave, delta_x2, 'C1-', linewidth=1, label=extname2)
            plot_reference_wavelengths(ax, refewave1, refewave2, extname1, extname2)
            ax.set_ylabel(r'$\Delta$x_ifu (pixel)')
        elif iplot == 1:
            ax.plot(wave, delta_y1, '.', label=extname1)
            if table2_adrcross is not None:
                ax.plot(wave, delta_y2, '-', linewidth=1, label=extname2)
            plot_reference_wavelengths(ax, refewave1, refewave2, extname1, extname2)
            ax.set_ylabel(r'$\Delta$y_ifu (pixel)')
        elif iplot == 2:
            ax.plot(wave, delta_x1 - delta_x2, '.')
            plot_reference_wavelengths(ax, refewave1, refewave2, extname1, extname2)
            ax.set_ylabel(r'difference in $\Delta$x_ifu (pixel)')
        else:
            ax.plot(wave, delta_y1 - delta_y2, '.')
            plot_reference_wavelengths(ax, refewave1, refewave2, extname1, extname2)
            ax.set_ylabel(r'difference in $\Delta$y_ifu (pixel)')
        ax.set_xlabel(f'Wavelength ({wave.unit})')
        ax.axhline(0, linestyle='--', color='gray')
        if iplot in [0, 1]:
            ax.legend()
    if suptitle is None:
        if extname2 is not None:
            plt.suptitle(f'file: {filename} (extensions: {extname1}, {extname2})')
        else:
            plt.suptitle(f'file: {filename} (extension: {extname1})')
    else:
        plt.suptitle(suptitle)
    plt.tight_layout()
    plt.show()


def main(args=None):
    # parse command-line options
    parser = argparse.ArgumentParser(description="Compare ADR extensions in 3D cube")
    parser.add_argument("filename", help="Input 3D FITS file")
    parser.add_argument("extname1", help="First extension name", type=str)
    parser.add_argument("extname2", help="Second extension name (optional)", type=str)
    parser.add_argument("--verbose", help="Display intermediate information", action="store_true")
    parser.add_argument("--echo", help="Display full command line", action="store_true")

    args = parser.parse_args(args=args)

    if len(sys.argv) == 1:
        parser.print_usage()
        raise SystemExit()

    if args.verbose:
        for arg, value in vars(args).items():
            print(ctext(f'{arg}: {value}', faint=True))

    if args.echo:
        print('\033[1m\033[31mExecuting: ' + ' '.join(sys.argv) + '\033[0m\n')

    for extname in [args.extname1, args.extname2]:
        if len(extname) > 8:
            raise ValueError(f"Extension '{extname}' must be less than 9 characters")

    compare_adr_extensions_in_3d_cube(
        filename=args.filename,
        extname1=args.extname1.upper(),
        extname2=args.extname2.upper()
    )


if __name__ == '__main__':
    main()




#
# Copyright 2024-2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Auxiliary class for linear wavelength calibration"""

import astropy.units as u


class LinearWaveCal(object):
    """Class to store a linear wavelength calibration.

    The parameters are stored making use of astropy units.

    Parameters
    ----------
    crpix1_wavecal : `~astropy.units.Quantity`
        CRPIX1 value of the wavelength calibrated spectrum.
    crval1_wavecal : `~astropy.units.Quantigy`
        CRVAL1 value of the wavelength calibrated spectrum.
    cdelt1_wavecal : `~astropy.units.Quantigy`
        CDELT1 value of the wavelength calibrated spectrum.
    naxis1_wavecal : `~astropy.units.Quantity`
        NAXIS1 value of the wavelength calibrated spectrum.
    default_wavelength_unit : `~astropy.units.core.Unit
        Default wavelength unit to be employed in CRVAL1 and CDELT1.

    Attributes
    ----------
    crpix1_wavecal : `~astropy.units.Quantity`
        CRPIX1 value of the wavelength calibrated spectrum.
    crval1_wavecal : `~astropy.units.Quantigy`
        CRVAL1 value of the wavelength calibrated spectrum.
    cdelt1_wavecal : `~astropy.units.Quantigy`
        CDELT1 value of the wavelength calibrated spectrum.
    naxis1_wavecal : `~astropy.units.Quantity`
        NAXIS1 value of the wavelength calibrated spectrum.
    default_wavelength_unit : `~astropy.units.core.Unit`
        Default wavelength unit to be employed in CRVAL1 and CDELT1.
    wmin : `~astropy.units.Quantity`
        Minimum wavelength, computed at pixel 0.5.
    wmax : `~astropy.units.Quantity`
        Maximum wavelength, computed at pixel naxis1_wavecal + 0.5.

    Methods
    -------
    wave_at_pixel(pixel):
        Compute wavelength(s) at the pixel coordinate(s).
    pixel_at_wave(wavelength)
        Compute pixel coordinate(s) at the wavelength(s).

    """

    def __init__(self, crpix1_wavecal, crval1_wavecal, cdelt1_wavecal, naxis1_wavecal, default_wavelength_unit):
        # check units
        if not crpix1_wavecal.unit.is_equivalent(u.pix):
            raise ValueError(f'Unexpected crpix1_wavecal unit: {crpix1_wavecal.unit}')
        if not crval1_wavecal.unit.is_equivalent(u.m):
            raise ValueError(f'Unexpected crval1_wavecal unit: {crval1_wavecal.unit}')
        if not cdelt1_wavecal.unit.is_equivalent(u.m / u.pix):
            raise ValueError(f'Unexpected cdelt1_wavecal unit: {cdelt1_wavecal.unit}')
        if not naxis1_wavecal.unit.is_equivalent(u.pix):
            raise ValueError(f'Unexpected naxis1_wavecal unit: {naxis1_wavecal.unit}')
        if not default_wavelength_unit.is_equivalent(u.m):
            raise ValueError(f'Unexpected default_wavelength_unit: {default_wavelength_unit}')

        # define attributes
        self.crpix1_wavecal = crpix1_wavecal
        self.crval1_wavecal = crval1_wavecal.to(default_wavelength_unit)
        self.cdelt1_wavecal = cdelt1_wavecal.to(default_wavelength_unit / u.pix)
        self.naxis1_wavecal = naxis1_wavecal
        self.wmin = self.wave_at_pixel(0.5 * u.pix)
        self.wmax = self.wave_at_pixel(naxis1_wavecal + 0.5 * u.pix)
        self.default_wavelength_unit = default_wavelength_unit

    def __str__(self):
        output = '<LinearWaveCal instance>\n'
        output += f'crpix1_wavecal: {self.crpix1_wavecal}\n'
        output += f'crval1_wavecal: {self.crval1_wavecal}\n'
        output += f'cdelt1_wavecal: {self.cdelt1_wavecal}\n'
        output += f'naxis1_wavecal: {self.naxis1_wavecal}\n'
        output += f'wmin..........: {self.wmin}\n'
        output += f'wmax..........: {self.wmax}\n'
        output += f'default_wavelength_unit: {self.default_wavelength_unit}'
        return output

    def __repr__(self):
        output = f'LinearWaveCal(\n'
        output += f'    crpix1_wavecal={self.crpix1_wavecal.value} * {self.crpix1_wavecal.unit.__repr__()},\n'
        output += f'    crval1_wavecal={self.crval1_wavecal.value} * {self.crval1_wavecal.unit.__repr__()},\n'
        output += f'    cdelt1_wavecal={self.cdelt1_wavecal.value} * {self.cdelt1_wavecal.unit.__repr__()},\n'
        output += f'    naxis1_wavecal={self.naxis1_wavecal.value} * {self.naxis1_wavecal.unit.__repr__()},\n'
        output += f'    default_wavelength_unit={self.default_wavelength_unit.__repr__()}\n'
        output += ')'
        return output

    def __eq__(self, other):
        if isinstance(other, LinearWaveCal):
            return self.__dict__ == other.__dict__
        else:
            return False

    def wave_at_pixel(self, pixel):
        """Compute wavelength(s) at the pixel coordinate(s).

        Parameters
        ----------
        pixel : `~astropy.units.Quantity`
            A single number or a numpy array with pixel coordinates.
            The units used serve to decide the criterion used to
            indicate the coordinates: u.pix for the FITS system
            (which starts at 1) and u.dimensionless_unscaled to
            indicate that the positions correspond to indices of
            an array (which starts at 0).

        Returns
        -------
        wave : `~astropy.units.Quantity`
            Wavelength computed at the considered pixel value(s).

        """

        if not isinstance(pixel, u.Quantity):
            raise ValueError(f"Object 'pixel': {pixel} is not a Quantity instance")

        if pixel.unit == u.pix:
            fitspixel = pixel
        elif pixel.unit == u.dimensionless_unscaled:
            fitspixel = (pixel.value + 1) * u.pix
        else:
            raise ValueError(f"Unexpected 'pixel' units: {pixel.unit}")

        wave = self.crval1_wavecal + (fitspixel - self.crpix1_wavecal) * self.cdelt1_wavecal

        return wave

    def pixel_at_wave(self, wave, return_units):
        """Compute pixel coordinate(s) at the wavelength(s).

        Parameters
        ----------
        wave : `~astropy.units.Quantity`
            A single number or a numpy array with wavelengths.
        return_units : `astropy.units.core.Unit`
            The return units serve to decide the criterion used to
            indicate the pixel coordinates: u.pix for the FITS system
            (which starts at 1) and u.dimensionless_unscaled to
            indicate that the positions correspond to indices of
            an array (which starts at 0).

        Returns
        -------
        pixel : `~astropy.units.Quantity`
            Pixel coordinates computed at the considered wavelength(s).

        """

        if not isinstance(wave, u.Quantity):
            raise ValueError(f"Object 'wave': {wave} is not a Quantity instance")
        if return_units not in [u.pix, u.dimensionless_unscaled]:
            raise ValueError(f'Unexpected return_units: {return_units}')

        waveunit = self.default_wavelength_unit
        fitspixel = (wave.to(waveunit) - self.crval1_wavecal) / self.cdelt1_wavecal + self.crpix1_wavecal

        if return_units == u.pix:
            return fitspixel
        else:
            return (fitspixel.value - 1) * u.dimensionless_unscaled



#
# Copyright 2008-2022 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import unittest

import numpy

from numina.array.nirproc import ramp_array


class FollowUpTheRampTestCase(unittest.TestCase):
    def setUp(self):
        self.ramp = numpy.empty((10, 1, 1))
        self.ti = 9.0
        self.sgain = 1.0
        self.sron = 1.0

    def test_exception(self):
        self.assertRaises(ValueError, ramp_array, self.ramp,
                          self.ti, -1.0, self.sron)
        self.assertRaises(ValueError, ramp_array, self.ramp,
                          self.ti, 0, self.sron)
        self.assertRaises(ValueError, ramp_array, self.ramp,
                          self.ti, self.sgain, -1.0)
        self.assertRaises(ValueError, ramp_array, self.ramp,
                          -1.0, self.sgain, self.sron)
        self.assertRaises(ValueError, ramp_array, self.ramp,
                          0.0, self.sgain, self.sron)
        self.assertRaises(ValueError, ramp_array, self.ramp,
                          self.ti, self.sgain, self.sron, saturation=-100)
        self.assertRaises(ValueError, ramp_array, self.ramp,
                          self.ti, self.sgain, self.sron, saturation=0)


class RampReadoutAxisTestCase(unittest.TestCase):

    def setUp(self):
        rows = 3
        columns = 4
        self.emptybp = numpy.zeros((rows, columns), dtype='uint8')
        self.data = numpy.arange(10, dtype='int32')
        self.data = numpy.tile(self.data, (columns, rows, 1)).T
        assert self.data.shape == (10, 3, 4)

        self.saturation = 65536
        self.ti = 9.0
        self.gain = 1.0
        self.ron = 1.0
        self.blank = 0

    def test_saturation0(self):
        '''Test we count correctly saturated pixels in RAMP mode.'''

        MASK_SATURATION = 3

        # No points
        self.data[:] = 50000  # - 32768
        saturation = 40000  # - 32767

        res = ramp_array(self.data, self.ti, self.gain, self.ron,
                         saturation=saturation,
                         blank=self.blank)

        for nn in res[2].flat:
            self.assertEqual(nn, 0)

        for n in res[3].flat:
            self.assertEqual(n, MASK_SATURATION)

        for v in res[1].flat:
            self.assertEqual(v, 0)

        for v in res[0].flat:
            self.assertEqual(v, self.blank)

    def test_saturation1(self):
        '''Test we count correctly saturated pixels in RAMP mode.'''

        MASK_GOOD = 0

        saturation = 50000
        self.data[7:, ...] = saturation

        res = ramp_array(self.data, self.ti, self.gain, self.ron,
                         saturation=saturation,
                         blank=self.blank)

        for nn in res[2].flat:
            self.assertEqual(nn, 7)

        for n in res[3].flat:
            self.assertEqual(n, MASK_GOOD)

        # for v in res[1].flat:
        #    self.assertEqual(v, 0.2142857142857143)

        for v in res[0].flat:
            self.assertAlmostEqual(v, 1)

    def test_dtypes0(self):
        '''Test output is float64 by default'''
        inttypes = ['int8', 'int16', 'int32', 'uint8', 'uint16', 'uint32']
        floattypes = ['float32', 'float64']
        if hasattr(numpy, 'float128'):
            floattypes.append('float128')
        mdtype = numpy.dtype('uint8')
        ddtype = numpy.dtype('float64')
        rows = 3
        columns = 4
        for dtype in inttypes:
            data = numpy.arange(10, dtype=dtype)
            data = numpy.tile(self.data, (columns, rows, 1)).T
            res = ramp_array(data, self.ti, self.gain, self.ron)
            self.assertIs(res[0].dtype, ddtype)
            self.assertIs(res[1].dtype, ddtype)
            self.assertIs(res[2].dtype, mdtype)
            self.assertIs(res[3].dtype, mdtype)

        for dtype in floattypes:
            data = numpy.arange(10, dtype=dtype)
            data = numpy.tile(self.data, (columns, rows, 1)).T
            res = ramp_array(data, self.ti, self.gain, self.ron)
            self.assertIs(res[0].dtype, ddtype)
            self.assertIs(res[1].dtype, ddtype)
            self.assertIs(res[2].dtype, mdtype)
            self.assertIs(res[3].dtype, mdtype)

    def test_badpixel0(self):
        '''Test we ignore badpixels in RAMP mode.'''
        self.emptybp[...] = 1

        res = ramp_array(self.data, self.ti, self.gain, self.ron,
                         saturation=self.saturation,
                         badpixels=self.emptybp,
                         blank=self.blank)

        for nn in res[2].flat:
            self.assertEqual(nn, 0)

        for n in res[3].flat:
            self.assertEqual(n, 1)

        for v in res[1].flat:
            self.assertEqual(v, 0)

        for v in res[0].flat:
            self.assertEqual(v, self.blank)

    def test_badpixel1(self):
        '''Test we handle correctly None badpixel mask.'''
        rows = 3
        columns = 4
        self.emptybp[...] = 0
        self.data = numpy.arange(10, dtype='int32')
        self.data = numpy.tile(self.data, (columns, rows, 1)).T
        value = 1.0
        variance = 0.13454545454545455

        # Badpixels is [0,..., 0]
        res = ramp_array(self.data, self.ti, self.gain, self.ron,
                         saturation=self.saturation,
                         badpixels=self.emptybp,
                         blank=self.blank)

        for n in res[3].flat:
            self.assertEqual(n, 0)

        for nn in res[2].flat:
            self.assertEqual(nn, 10)

        # for v in res[1].flat:
        #    self.assertAlmostEqual(v, variance)

        for v in res[0].flat:
            self.assertAlmostEqual(v, value)

        # Badpixels is None
        res = ramp_array(self.data, self.ti, self.gain, self.ron,
                         saturation=self.saturation,
                         badpixels=None,
                         blank=self.blank)

        # for n in res[4].flat:
        #    self.assertEqual(n, 0)

        for n in res[3].flat:
            self.assertEqual(n, 0)

        for nn in res[2].flat:
            self.assertEqual(nn, 10)

        for v in res[1].flat:
            self.assertAlmostEqual(v, variance)

        for v in res[0].flat:
            self.assertAlmostEqual(v, value)

        # Badpixels has default value
        res = ramp_array(self.data, self.ti, self.gain, self.ron,
                         saturation=self.saturation,
                         blank=self.blank)

        # for n in res[4].flat:
        #    self.assertEqual(n, 0)

        for n in res[3].flat:
            self.assertEqual(n, 0)

        for nn in res[2].flat:
            self.assertEqual(nn, 10)

        for v in res[1].flat:
            self.assertAlmostEqual(v, variance)

        for v in res[0].flat:
            self.assertAlmostEqual(v, value)

    def test_results1(self):
        '''Test we obtain correct values in RAMP mode'''

        res = ramp_array(self.data, self.ti, self.gain, self.ron,
                         saturation=self.saturation,
                         blank=self.blank)

        for nn in res[2].flat:
            self.assertEqual(nn, 10)

        for n in res[3].flat:
            self.assertEqual(n, 0)

#        for v in res[1].flat:
#            self.assertEqual(v, 0.13454545454545455)

        for v in res[0].flat:
            self.assertEqual(v, 1.0)

    def test_results2(self):
        '''Test we obtain correct values in RAMP mode'''

        self.data *= 12
        self.data[:, 1, 1] = 70000
        # self.data[5:,2,2] += 1300

        self.emptybp[0, 0] = 1

        res = ramp_array(self.data, self.ti, self.gain, self.ron,
                         dtype='float32',
                         saturation=self.saturation,
                         badpixels=self.emptybp,
                         blank=self.blank)

        res0 = 12 * numpy.ones((3, 4), dtype='float32')
        res0[0, 0] = 0
        res0[1, 1] = 0
        res1 = 1.4812121212 * numpy.ones((3, 4), dtype='float32')
        res1[0, 0] = 0
        res1[1, 1] = 0
        res1[2, 2] = 1.61
        res2 = 10 * numpy.ones((3, 4), dtype='uint8')
        res2[0, 0] = 0
        res2[1, 1] = 0
        res3 = numpy.zeros((3, 4), dtype='uint8')
        res3[0, 0] = 1
        res3[1, 1] = 3
        res4 = numpy.zeros((3, 4), dtype='uint8')
        res4[2, 2] = 5

        for xx, yy in zip(res[0].flat, res0.flat):
            self.assertAlmostEqual(xx, yy)
        # for xx, yy in zip(res[1].flat, res1.flat):
        #    self.assertAlmostEqual(xx, yy)
        for xx, yy in zip(res[2].flat, res2.flat):
            self.assertEqual(xx, yy)
        for xx, yy in zip(res[3].flat, res3.flat):
            self.assertEqual(xx, yy)
        # for xx, yy in zip(res[4].flat, res4.flat):
        #    self.assertEqual(xx, yy)




import numpy
import numpy as np

import numina.array._combine as _c  # noqa


def test_error2():
    """test general functionality"""
    data1 = numpy.array([[-1, -3], [5, 0]])
    data2 = numpy.array([[1, 2], [8, 3]])
    data3 = numpy.array([[9, 1], [2, 4]])
    mask1 = numpy.array([[0, 0], [0, 0]])
    mask2 = numpy.array([[0, 0], [0, 1]])
    mask3 = numpy.array([[0, 0], [0, 1]])
    exp_res = numpy.array([[3, 0], [5, 0]])
    exp_var = numpy.array([[28, 7], [9, 0]])
    exp_pix = numpy.array([[3, 3], [3, 1]])
    a, b, c = _c.generic_combine_miter(_c.mean_method(), [data1, data2, data3], masks=[mask1, mask2, mask3])
    assert np.allclose(a, exp_res)
    assert np.allclose(b, exp_var)
    assert np.allclose(c, exp_pix)


def test_error3():
    """test general functionality"""
    data1 = numpy.array([[-1, -3], [5, 0]])
    data2 = numpy.array([[1, 2], [8, 3]])
    data3 = numpy.array([[9, 1], [2, 4]])
    mask1 = numpy.array([[0, 0], [0, 0]])
    mask2 = numpy.array([[0, 0], [0, 0]])
    mask3 = numpy.array([[0, 0], [0, 0]])
    out_res = numpy.array([[1, 2], [1, 2]])
    out_var = numpy.array([[1, 2], [1, 2]])
    out_pix = numpy.array([[1, 2], [1, 2]])
    exp_res = numpy.array([[3, 0], [5, 2]])
    exp_var = numpy.array([[28, 7], [9, 4]])
    exp_pix = numpy.array([[3, 3], [3, 3]])
    a, b, c = _c.generic_combine_miter(_c.mean_method(), [data1, data2, data3], masks=[mask1, mask2, mask3],
                                       out_res=out_res, out_var=out_var, out_pix=out_pix)
    assert np.allclose(a, exp_res)
    assert np.allclose(b, exp_var)
    assert np.allclose(c, exp_pix)


def test_error4():
    """test general functionality"""
    data1 = numpy.array([[-1, -3], [5, 0]])
    out_res = numpy.array([[1, 2], [1, 2]])
    out_var = numpy.array([[1, 2], [1, 2]])
    out_pix = numpy.array([[1, 2], [1, 2]])
    exp_res = numpy.array([[-1, -3], [5, 0]])
    exp_var = numpy.array([[0, 0], [0, 0]])
    exp_pix = 100 * numpy.array([[1, 1], [1, 1]])
    a, b, c = _c.generic_combine_miter(_c.mean_method(), [data1]*100,
                                       out_res=out_res, out_var=out_var, out_pix=out_pix)
    assert np.allclose(a, exp_res)
    assert np.allclose(b, exp_var)
    assert np.allclose(c, exp_pix)


def test_error5():
    """test general functionality"""
    data1 = numpy.array([[-1, -3], [5, 0]])
    exp_res = numpy.array([[-1, -3], [5, 0]])
    exp_var = numpy.array([[0, 0], [0, 0]])
    exp_pix = 100 * numpy.array([[1, 1], [1, 1]])
    a, b, c = _c.generic_combine_miter(_c.mean_method(), [data1]*100)
    assert np.allclose(a, exp_res)
    assert np.allclose(b, exp_var)
    assert np.allclose(c, exp_pix)


def test_error6():
    """test general functionality"""
    data1 = numpy.array([[-1, -3], [5, 0]])
    data2 = numpy.array([[2, 0], [4, 1]])
    exp_res = numpy.array([[0.5, -1.5], [4.5, 0.5]])
    exp_var = numpy.array([[4.5, 4.5], [0.5, 0.5]])
    exp_pix = 2 * numpy.array([[1, 1], [1, 1]])
    a, b, c = _c.generic_combine_miter(_c.mean_method(), [data1, data2])
    assert np.allclose(a, exp_res)
    assert np.allclose(b, exp_var)
    assert np.allclose(c, exp_pix)



#
# Copyright 2014-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""FITS header schema and validation.

This module is a simplification of the FITS Schema defined
by Erik Bray here:
http://embray.github.io/PyFITS/schema/users_guide/users_schema.html

If this schema implementation reaches pyfits/astropy stable,
we will use it instead of ours, with schema definitions
being the same.

"""


class SchemaValidationError(Exception):
    """Exception raised when a Schema does not validate a FITS header."""
    pass


class SchemaDefinitionError(Exception):
    """Exception raised when a FITS Schema definition is not valid."""
    pass


def _from_ipt(value):
    if isinstance(value, bool):
        return (value, bool)
    if isinstance(value, str):
        return (value, str)
    elif isinstance(value, int):
        return (value, int)
    elif isinstance(value, float):
        return (value, float)
    elif isinstance(value, complex):
        return (value, complex)
    elif value in [str, int, float, complex, bool]:
        return (None, value)
    elif isinstance(value, list):
        if value:
            _, type_ = _from_ipt(value[0])
            return value, type_
        else:
            raise SchemaDefinitionError(value)
    else:
        raise SchemaDefinitionError(value)


class SchemaKeyword(object):
    """A keyword in the schema"""

    def __init__(self, name, mandatory=False, valid=True,
                 value=None):
        self.name = name
        self.mandatory = mandatory
        self.valid = valid
        if self.mandatory and not self.valid:
            raise SchemaDefinitionError(
                "keyword 'cannot be 'mandatory' and "
                "'not valid'"
            )
        self.choose = False
        self.valcheck = False
        self.value = None
        self.type_ = None
        if value is not None:
            self.value, self.type_ = _from_ipt(value)
            if self.value is not None:
                self.valcheck = True
                if isinstance(self.value, list):
                    self.choose = True

    def validate(self, header):
        sname = 'schema'
        # check the keyword is defined
        val = header.get(self.name)

        if val is None:
            if self.mandatory:
                raise SchemaValidationError(
                    sname, f'mandatory keyword {self.name!r} missing from header')

            # In the rest of cases
            return True
        else:
            if not self.valid:
                raise SchemaValidationError(
                    sname, f'invalid keyword {self.name!r} present in header')

        # Cases here
        # val is not None and key id mandatory or valid

        if not self.type_:
            # We dont have type information
            # Nothing more to do
            return True
        else:
            if not isinstance(val, self.type_):
                raise SchemaValidationError(
                    sname, 'keyword %r is required to have a value of type %r'
                    '; got a value of type %r instead' %
                    (self.name, self.type_.__name__, type(val).__name__))
            # Check value
            if self.choose:
                if val not in self.value:
                    raise SchemaValidationError(
                        sname,
                        'keyword %r is required to have one of the values %r; '
                        'got %r instead' %
                        (self.name, self.value, val))
                else:
                    return True
            elif self.valcheck:
                if val != self.value:
                    raise SchemaValidationError(
                        sname,
                        'keyword %r is required to have the value %r; got '
                        '%r instead' % (self.name, self.value, val))
            else:
                pass

        return True


class Schema(object):
    """A FITS schema"""

    def __init__(self, sc):
        self.kwl = []
        self.extend(sc)

    def validate(self, header):
        for ll in self.kwl:
            ll.validate(header)

    def extend(self, sc):
        kw = sc.get('keywords', {})
        for k, v in kw.items():
            mandatory = v.get('mandatory', False)
            valid = v.get('valid', True)
            value = v.get('value', None)
            sk = SchemaKeyword(
                k, mandatory=mandatory,
                valid=valid, value=value
            )
            self.kwl.append(sk)


types_table = {
    'string': str,
    'number': float,
    'integer': int,
    'bool': bool
}


class SchemaNode(object):
    def __init__(self, *args, **kwargs):
        self.title = kwargs.get('title', 'undefined')
        self.required = kwargs.get('required', True)
        #
        self.group = kwargs.get('group', None)
        self.matched = kwargs.get('matched', None)
        #
        self.children_obj = {}

    def validate_req(self, value, state):
        if self.group is not None:
            print('this node has group', self.group)
        # value must be hdulist
        pass

    def validate(self, value):
        pass


class SchemaExtension(SchemaNode):
    def __init__(self, *args, **kwargs):
        super(SchemaExtension, self).__init__(*args, **kwargs)

    def validate(self, value):
        # value must be hdulist
        pass

    def validate_req(self, value, state):
        # super(SchemaExtension, self).validate_req(value, state)

        if self.group is not None:
            print('this node has group', self.group)
            print('matched', self.matched)
            if self.group in state['groups']:
                # both must match
                if state['groups'][self.group] != self.matched:
                    msg = 'matched group {} differs from state group {}'
                    raise ValueError(msg.format(
                        self.matched, state['groups'][self.group]))

        for key, val in self.children_obj.items():
            val.validate_req(value, state)
            print(state)


class SchemaKeywordII(SchemaNode):
    def __init__(self, *args, **kwargs):
        super(SchemaKeywordII, self).__init__(*args, **kwargs)

        self.v_type = kwargs.get('type', None)
        self.enum_t = kwargs.get('enum', None)
        self.value_t = kwargs.get('value', None)

        # if enum_t is a list, and group is set
        # matched must be a list of the same length

    def validate_req(self, value, state):
        print('validate keyword', self.title)

        hvalue = value.get(self.title)
        matched_val = None
        print('hvalue', hvalue)
        sname = 'test'
        key = 0
        if self.required:
            print('keyword required')
            # if not defined, fail
            if hvalue is None:
                raise SchemaValidationError(
                    sname, f'required keyword {key!r} missing from header')
        else:
            print('keyword not required')
        # check type
        if self.v_type:
            ptype = types_table[self.v_type]
            if not isinstance(hvalue, ptype):
                raise SchemaValidationError(
                    sname, 'keyword %r is required to have a value of type %r'
                           '; got a value of type %r instead' %
                           (key, ptype.__name__, type(hvalue).__name__))

        if self.enum_t:
            try:
                match = self.enum_t.index(hvalue)
                matched_val = self.matched[match]

            except ValueError:
                raise SchemaValidationError(
                    sname,
                    'keyword %r is required to have one of the values %r; '
                    'got %r instead' %
                    (self.title, self.enum_t, hvalue))

        if self.value_t:
            matched_val = self.matched
            if self.value_t != hvalue:
                raise SchemaValidationError(
                    sname,
                    'keyword %r is required to have value %r; '
                    'got %r instead' %
                    (key, self.value_t, hvalue))

        if self.group is not None:
            print('this node has group', self.group)
            print('matched', self.matched)
            if self.group in state['groups']:
                # both must match
                if state['groups'][self.group] != matched_val:
                    msg = 'matched group {} differs from state group {}'
                    raise ValueError(msg.format(
                        matched_val, state['groups'][self.group]))
                else:
                    msg = 'matched group {} = value from state group {}'
                    print(msg.format(matched_val, state['groups'][self.group]))
            else:
                state['groups'][self.group] = matched_val


def create_keyword_node(node, key=None):
    if 'title' not in node:
        node['title'] = key
    return SchemaKeywordII(**node)


def create_extension_node(node):
    ext = SchemaExtension(**node)
    keywords = node.get('keywords', {})
    for k, knode in keywords.items():
        ext.children_obj[k] = create_keyword_node(knode, key=k)
    return ext


class SchemeKeyword(object):
    def __init__(self, *args, **kwargs):
        types_table = {
            'string': str,
            'number': float,
            'integer': int,
            'bool': bool
        }

        self.key = "key"
        self.required = kwargs.get('required', True)
        self.enum = kwargs.get('enum', None)
        self.value = kwargs.get('value', None)
        self.type = kwargs.get('type', None)
        self.type_p = types_table.get(self.type, None)

    def validate(self, value):
        sname = 'titlw'
        key = "ky"
        if self.required:
            if value is None:
                msg = f'required keyword {key!r} missing from header'
                raise SchemaValidationError(sname, msg)
        if self.type:
            if not isinstance(value, self.type_p):
                raise SchemaValidationError(
                    sname, 'keyword %r is required to have a value of type %r'
                           '; got a value of type %r instead' %
                           (key, self.type_p.__name__, type(value).__name__))


class SchemeKeywordString(SchemeKeyword):
    def __init__(self, *args, **kwargs):
        super(SchemeKeywordString, self).__init__(*args, **kwargs)


def validate(header, schema):
    sname = schema.get('title', 'schema')
    schema_keys = schema.get('keywords', {})
    for key, desc in schema_keys.items():
        required = desc.get('required', True)
        vtype = desc.get('type', None)
        enumt = desc.get('enum', None)
        valuet = desc.get('value', None)
        hvalue = header.get(key, None)
        if required:
            # if not defined, fail
            if hvalue is None:
                raise SchemaValidationError(
                    sname, f'required keyword {key!r} missing from header')
        # check type
        if vtype:
            ptype = types_table[vtype]
            if not isinstance(hvalue, ptype):
                raise SchemaValidationError(
                    sname, 'keyword %r is required to have a value of type %r'
                           '; got a value of type %r instead' %
                           (key, ptype.__name__, type(hvalue).__name__))

        if enumt:
            if hvalue not in enumt:
                raise SchemaValidationError(
                    sname,
                    'keyword %r is required to have one of the values %r; '
                    'got %r instead' %
                    (hvalue, enumt, hvalue))

        if valuet:
            if valuet != hvalue:
                raise SchemaValidationError(
                    sname,
                    'keyword %r is required to have value %r; '
                    'got %r instead' %
                    (key, valuet, hvalue))


def validate_image(image, schema):
    print('validate image')
    state = {}
    state['groups'] = {}

    for extname, extnode in schema['extensions'].items():

        node = create_extension_node(extnode)
        if extname in image:
            node.validate_req(image[extname].header, state)
        else:
            if node.required:
                print('required ext', extname, 'not present')
            else:
                print('not required ext', extname, 'not present')



#
# Copyright 2015-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Supervised quality control of wavelength calibration"""

import argparse
import astropy.io.fits as fits
import decimal
import numpy as np
import os

from numina.array.stats import summary
from numina.array.display.iofunctions import readi
from numina.array.display.iofunctions import readf
from numina.array.display.iofunctions import readc
from numina.array.display.matplotlib_qt import set_window_geometry
from numina.array.display.pause_debugplot import pause_debugplot
from numina.array.display.polfit_residuals import polfit_residuals
from numina.array.display.polfit_residuals \
    import polfit_residuals_with_sigma_rejection
from .peaks_spectrum import find_peaks_spectrum
from .peaks_spectrum import refine_peaks_spectrum

from numina.array.display.pause_debugplot import DEBUGPLOT_CODES


def match_wv_arrays(wv_master, wv_expected_all_peaks, delta_wv_max):
    """Verify expected wavelength for each line peak.

    Assign individual arc lines from wv_master to each expected
    wavelength when the latter is within the maximum allowed range.

    Parameters
    ----------
    wv_master : numpy array
        Array containing the master wavelengths.
    wv_expected_all_peaks : numpy array
        Array containing the expected wavelengths computed from the
        approximate polynomial calibration applied to the location of
        the line peaks.
    delta_wv_max : float
        Maximum distance to accept that the master wavelength
        corresponds to the expected wavelength.

    Returns
    -------
    wv_verified_all_peaks : numpy array
        Verified wavelengths from master list.

    """

    # initialize the output array to zero
    wv_verified_all_peaks = np.zeros_like(wv_expected_all_peaks)

    # initialize to True array to indicate that no peak has already
    # been verified (this flag avoids duplication)
    wv_unused = np.ones_like(wv_expected_all_peaks, dtype=bool)

    # since it is likely that len(wv_master) < len(wv_expected_all_peaks),
    # it is more convenient to execute the search in the following order
    for i in range(len(wv_master)):
        j = np.searchsorted(wv_expected_all_peaks, wv_master[i])
        if j == 0:
            if wv_unused[j]:
                delta_wv = abs(wv_master[i] - wv_expected_all_peaks[j])
                if delta_wv < delta_wv_max:
                    wv_verified_all_peaks[j] = wv_master[i]
                    wv_unused[j] = False
        elif j == len(wv_expected_all_peaks):
            if wv_unused[j-1]:
                delta_wv = abs(wv_master[i] - wv_expected_all_peaks[j-1])
                if delta_wv < delta_wv_max:
                    wv_verified_all_peaks[j-1] = wv_master[i]
                    wv_unused[j-1] = False
        else:
            delta_wv1 = abs(wv_master[i] - wv_expected_all_peaks[j-1])
            delta_wv2 = abs(wv_master[i] - wv_expected_all_peaks[j])
            if delta_wv1 < delta_wv2:
                if delta_wv1 < delta_wv_max:
                    if wv_unused[j-1]:
                        wv_verified_all_peaks[j-1] = wv_master[i]
                        wv_unused[j-1] = False
                    elif wv_unused[j]:
                        if delta_wv2 < delta_wv_max:
                            wv_verified_all_peaks[j] = wv_master[i]
                            wv_unused[j] = False
            else:
                if delta_wv2 < delta_wv_max:
                    if wv_unused[j]:
                        wv_verified_all_peaks[j] = wv_master[i]
                        wv_unused[j] = False
                    elif wv_unused[j-1]:
                        if delta_wv1 < delta_wv_max:
                            wv_verified_all_peaks[j-1] = wv_master[i]
                            wv_unused[j-1] = False

    return wv_verified_all_peaks


def fun_wv(xchannel, crpix1, crval1, cdelt1):
    """Compute wavelengths from channels.

    The wavelength calibration is provided through the usual parameters
    CRPIX1, CRVAL1 and CDELT1.

    Parameters
    ----------
    xchannel : numpy array
        Input channels where the wavelengths will be evaluated.
    crpix1: float
        CRPIX1 keyword.
    crval1: float
        CRVAL1 keyword.
    cdelt1: float
        CDELT1 keyword.

    Returns
    -------
    wv : numpy array
        Computed wavelengths

    """
    wv = crval1 + (xchannel - crpix1) * cdelt1
    return wv


def check_wlcalib_sp(sp, crpix1, crval1, cdelt1, wv_master,
                     coeff_ini=None, naxis1_ini=None,
                     min_nlines_to_refine=0,
                     interactive=False,
                     threshold=0,
                     nwinwidth_initial=7,
                     nwinwidth_refined=5,
                     ntimes_match_wv=2,
                     poldeg_residuals=1,
                     times_sigma_reject=5,
                     use_r=False,
                     title=None,
                     remove_null_borders=True,
                     ylogscale=False,
                     geometry=None,
                     pdf=None,
                     debugplot=0):
    """Check wavelength calibration of the provided spectrum.

    Parameters
    ----------
    sp : numpy array
        Wavelength calibrated spectrum.
    crpix1: float
        CRPIX1 keyword.
    crval1: float
        CRVAL1 keyword.
    cdelt1: float
        CDELT1 keyword.
    wv_master: numpy array
        Array with the detailed list of expected arc lines.
    coeff_ini : array like
        Coefficients initially employed to obtain the wavelength
        calibration of the provided spectrum. When this coefficients
        are provided, this function computes a refined version of
        them, incorporating the corrections derived from the fit to
        the residuals.
    naxis1_ini : int
        NAXIS1 in original spectrum employed to fit the initial
        wavelength calibration.
    min_nlines_to_refine : int
        Minimum number of identified lines necessary to perform the
        wavelength calibration refinement. If zero, no minimum number
        is required.
    interactive : bool
        If True, the function allows the user to modify the residuals
        fit.
    threshold : float
        Minimum signal in the peaks.
    nwinwidth_initial : int
        Width of the window where each peak must be initially found.
    nwinwidth_refined : int
        Width of the window where each peak must be refined.
    ntimes_match_wv : float
        Times CDELT1 to match measured and expected wavelengths.
    poldeg_residuals : int
        Polynomial degree for fit to residuals.
    times_sigma_reject : float or None
        Number of times the standard deviation to reject points
        iteratively. If None, the fit does not reject any point.
    use_r : bool
        If True, additional statistical analysis is performed using R.
    title : string
        Plot title.
    remove_null_borders : bool
        If True, remove leading and trailing zeros in spectrum.
    ylogscale : bool
        If True, the spectrum is displayed in logarithmic units. Note
        that this is only employed for display purposes. The line peaks
        are found in the original spectrum.
    geometry : tuple (4 integers) or None
        x, y, dx, dy values employed to set the window geometry.
    pdf : PdfFile object or None
        If not None, output is sent to PDF file.
    debugplot : int
        Debugging level for messages and plots. For details see
        'numina.array.display.pause_debugplot.py'.

    Returns
    -------
    coeff_refined : numpy array
        Refined version of the initial wavelength calibration
        coefficients. These coefficients are computed only when
        the input parameter 'coeff_ini' is not None.

    """

    # protections
    if type(sp) is not np.ndarray:
        raise ValueError("sp must be a numpy.ndarray")
    elif sp.ndim != 1:
        raise ValueError("sp.ndim is not 1")

    if coeff_ini is None and naxis1_ini is None:
        pass
    elif coeff_ini is not None and naxis1_ini is not None:
        pass
    else:
        raise ValueError("coeff_ini and naxis1_ini must be simultaneously "
                         "None of both different from None")

    # check that interactive use takes place when plotting
    if interactive:
        if abs(debugplot) % 10 == 0:
            raise ValueError("ERROR: interative use of this function is not "
                             "possible when debugplot=", debugplot)

    # interactive and pdf are incompatible
    if interactive:
        if pdf is not None:
            raise ValueError("ERROR: interactive use of this function is not "
                             "possible when pdf is not None")

    # display list of expected arc lines
    if abs(debugplot) in (21, 22):
        print('wv_master:', wv_master)

    # determine spectrum length
    naxis1 = sp.shape[0]

    # define default values in case no useful lines are identified
    fxpeaks = np.array([])
    ixpeaks_wv = np.array([])
    fxpeaks_wv = np.array([])
    wv_verified_all_peaks = np.array([])
    nlines_ok = 0
    xresid = np.array([], dtype=float)
    yresid = np.array([], dtype=float)
    reject = np.array([], dtype=bool)
    polyres = np.polynomial.Polynomial([0])
    poldeg_effective = 0
    ysummary = summary(np.array([]))
    local_ylogscale = ylogscale

    # find initial line peaks
    ixpeaks = find_peaks_spectrum(sp,
                                  nwinwidth=nwinwidth_initial,
                                  threshold=threshold)
    npeaks = len(ixpeaks)

    if npeaks > 0:
        # refine location of line peaks
        fxpeaks, sxpeaks = refine_peaks_spectrum(
            sp, ixpeaks,
            nwinwidth=nwinwidth_refined,
            method="gaussian"
        )
        ixpeaks_wv = fun_wv(ixpeaks + 1, crpix1, crval1, cdelt1)
        fxpeaks_wv = fun_wv(fxpeaks + 1, crpix1, crval1, cdelt1)

        # match peaks with expected arc lines
        delta_wv_max = ntimes_match_wv * cdelt1
        wv_verified_all_peaks = match_wv_arrays(
            wv_master,
            fxpeaks_wv,
            delta_wv_max=delta_wv_max
        )

    loop = True

    while loop:

        if npeaks > 0:

            lines_ok = np.where(wv_verified_all_peaks > 0)
            nlines_ok = len(lines_ok[0])

            # there are matched lines
            if nlines_ok > 0:
                # compute residuals
                xresid = fxpeaks_wv[lines_ok]
                yresid = wv_verified_all_peaks[lines_ok] - fxpeaks_wv[lines_ok]

                # determine effective polynomial degree
                if nlines_ok > poldeg_residuals:
                    poldeg_effective = poldeg_residuals
                else:
                    poldeg_effective = nlines_ok - 1

                # fit polynomial to residuals
                polyres, yresres, reject = \
                    polfit_residuals_with_sigma_rejection(
                        x=xresid,
                        y=yresid,
                        deg=poldeg_effective,
                        times_sigma_reject=times_sigma_reject,
                        use_r=use_r,
                        debugplot=0
                    )
                ysummary = summary(yresres)
            else:
                polyres = np.polynomial.Polynomial([0.0])

        list_wv_found = [str(round(wv, 4))
                         for wv in wv_verified_all_peaks if wv != 0]
        list_wv_master = [str(round(wv, 4)) for wv in wv_master]
        set1 = set(list_wv_master)
        set2 = set(list_wv_found)
        missing_wv = list(set1.symmetric_difference(set2))
        missing_wv.sort()
        if abs(debugplot) >= 10:
            print('-' * 79)
            print(">>> Number of arc lines in master file:", len(wv_master))
        if abs(debugplot) in [21, 22]:
            print(">>> Unmatched lines...................:", missing_wv)
        elif abs(debugplot) >= 10:
            print(">>> Number of unmatched lines.........:", len(missing_wv))
        if abs(debugplot) >= 10:
            print(">>> Number of line peaks found........:", npeaks)
            print(">>> Number of identified lines........:", nlines_ok)
            print(">>> Number of unmatched lines.........:", len(missing_wv))
            print(">>> Polynomial degree in residuals fit:", poldeg_effective)
            print(">>> Polynomial fit to residuals.......:\n", polyres)

        # display results
        if (abs(debugplot) % 10 != 0) or (pdf is not None):
            from numina.array.display.matplotlib_qt import plt
            if pdf is not None:
                fig = plt.figure(figsize=(11.69, 8.27), dpi=100)
            else:
                fig = plt.figure()
            set_window_geometry(geometry)

            # residuals
            ax2 = fig.add_subplot(2, 1, 1)
            if nlines_ok > 0:
                ymin = min(yresid)
                ymax = max(yresid)
                dy = ymax - ymin
                if dy > 0:
                    ymin -= dy/20
                    ymax += dy/20
                else:
                    ymin -= 0.5
                    ymax += 0.5
            else:
                ymin = -1.0
                ymax = 1.0
            ax2.set_ylim(ymin, ymax)
            if nlines_ok > 0:
                ax2.plot(xresid, yresid, 'o')
                ax2.plot(xresid[reject], yresid[reject], 'o', color='tab:gray')
            ax2.set_ylabel('Offset ' + r'($\AA$)')
            ax2.yaxis.label.set_size(10)
            if title is not None:
                ax2.set_title(title, **{'size': 12})
            xwv = fun_wv(np.arange(naxis1) + 1.0, crpix1, crval1, cdelt1)
            ax2.plot(xwv, polyres(xwv), '-')
            ax2.text(1, 0, 'CDELT1 (' + r'$\AA$' + '/pixel)=' + str(cdelt1),
                     horizontalalignment='right',
                     verticalalignment='bottom',
                     transform=ax2.transAxes)
            ax2.text(0, 0, 'Wavelength ' + r'($\AA$) --->',
                     horizontalalignment='left',
                     verticalalignment='bottom',
                     transform=ax2.transAxes)
            ax2.text(0, 1, 'median=' +
                     str(round(ysummary['median'], 4)) + r' $\AA$',
                     horizontalalignment='left',
                     verticalalignment='top',
                     transform=ax2.transAxes)
            ax2.text(0.5, 1, 'npoints (total / used / removed)',
                     horizontalalignment='center',
                     verticalalignment='top',
                     transform=ax2.transAxes)
            ax2.text(0.5, 0.92,
                     str(ysummary['npoints']) + ' / ' +
                     str(ysummary['npoints'] - sum(reject)) + ' / ' +
                     str(sum(reject)),
                     horizontalalignment='center',
                     verticalalignment='top',
                     transform=ax2.transAxes)
            ax2.text(1, 1, 'robust_std=' +
                     str(round(ysummary['robust_std'], 4)) + r' $\AA$',
                     horizontalalignment='right',
                     verticalalignment='top',
                     transform=ax2.transAxes)

            # median spectrum and peaks
            # remove leading and trailing zeros in spectrum when requested
            if remove_null_borders:
                nonzero = np.nonzero(sp)[0]
                j1 = nonzero[0]
                j2 = nonzero[-1]
                xmin = xwv[j1]
                xmax = xwv[j2]
            else:
                xmin = min(xwv)
                xmax = max(xwv)
            dx = xmax - xmin
            if dx > 0:
                xmin -= dx / 80
                xmax += dx / 80
            else:
                xmin -= 0.5
                xmax += 0.5

            if local_ylogscale:
                spectrum = sp - sp.min() + 1.0
                spectrum = np.log10(spectrum)
                ymin = spectrum[ixpeaks].min()
            else:
                spectrum = sp.copy()
                ymin = min(spectrum)
            ymax = max(spectrum)
            dy = ymax - ymin
            if dy > 0:
                ymin -= dy/20
                ymax += dy/20
            else:
                ymin -= 0.5
                ymax += 0.5
            ax1 = fig.add_subplot(2, 1, 2, sharex=ax2)
            ax1.set_xlim(xmin, xmax)
            ax1.set_ylim(ymin, ymax)
            ax1.plot(xwv, spectrum)
            if npeaks > 0:
                ax1.plot(ixpeaks_wv, spectrum[ixpeaks], 'o',
                         fillstyle='none', label="initial location")
                ax1.plot(fxpeaks_wv, spectrum[ixpeaks], 'o',
                         fillstyle='none', label="refined location")
                lok = wv_verified_all_peaks > 0
                ax1.plot(fxpeaks_wv[lok], spectrum[ixpeaks][lok], 'go',
                         label="valid line")
            if local_ylogscale:
                ax1.set_ylabel('~ log10(number of counts)')
            else:
                ax1.set_ylabel('number of counts')
            ax1.yaxis.label.set_size(10)
            ax1.xaxis.tick_top()
            ax1.xaxis.set_label_position('top')
            for i in range(len(ixpeaks)):
                # identified lines
                if wv_verified_all_peaks[i] > 0:
                    ax1.text(fxpeaks_wv[i], spectrum[ixpeaks[i]],
                             str(wv_verified_all_peaks[i]) +
                             '(' + str(i + 1) + ')',
                             fontsize=8,
                             horizontalalignment='center')
                else:
                    ax1.text(fxpeaks_wv[i], spectrum[ixpeaks[i]],
                             '(' + str(i + 1) + ')',
                             fontsize=8,
                             horizontalalignment='center')
                # estimated wavelength from initial calibration
                if npeaks > 0:
                    estimated_wv = fun_wv(fxpeaks[i] + 1,
                                          crpix1, crval1, cdelt1)
                    estimated_wv = str(round(estimated_wv, 4))
                    ax1.text(fxpeaks_wv[i], ymin,  # spmedian[ixpeaks[i]],
                             estimated_wv, fontsize=8, color='grey',
                             rotation='vertical',
                             horizontalalignment='center',
                             verticalalignment='top')
            if len(missing_wv) > 0:
                tmp = [float(wv) for wv in missing_wv]
                ax1.vlines(tmp, ymin=ymin, ymax=ymax,
                           colors='grey', linestyles='dotted',
                           label='missing lines')
            ax1.legend()
            if pdf is not None:
                pdf.savefig()
            else:
                if debugplot in [-22, -12, 12, 22]:
                    pause_debugplot(
                        debugplot=debugplot,
                        optional_prompt='Zoom/Unzoom or ' +
                                        'press RETURN to continue...',
                        pltshow=True
                    )
                else:
                    pause_debugplot(debugplot=debugplot, pltshow=True)

            # display results and request next action
            if interactive:
                print('Recalibration menu')
                print('------------------')
                print('[d] (d)elete all the identified lines')
                print('[r] (r)estart from begining')
                print('[a] (a)utomatic line inclusion')
                print('[l] toggle (l)ogarithmic scale on/off')
                print('[p] modify (p)olynomial degree')
                print('[o] (o)utput data with identified line peaks')
                print('[x] e(x)it without additional changes')
                print('[#] from 1 to ' + str(len(ixpeaks)) +
                      ' --> modify line #')
                ioption = readi('Option', default='x',
                                minval=1, maxval=len(ixpeaks),
                                allowed_single_chars='adloprx')
                if ioption == 'd':
                    wv_verified_all_peaks = np.zeros(npeaks)
                elif ioption == 'r':
                    delta_wv_max = ntimes_match_wv * cdelt1
                    wv_verified_all_peaks = match_wv_arrays(
                        wv_master,
                        fxpeaks_wv,
                        delta_wv_max=delta_wv_max
                    )
                elif ioption == 'a':
                    fxpeaks_wv_corrected = np.zeros_like(fxpeaks_wv)
                    for i in range(npeaks):
                        fxpeaks_wv_corrected[i] = fxpeaks_wv[i] + \
                            polyres(fxpeaks_wv[i])
                    delta_wv_max = ntimes_match_wv * cdelt1
                    wv_verified_all_peaks = match_wv_arrays(
                        wv_master,
                        fxpeaks_wv_corrected,
                        delta_wv_max=delta_wv_max
                    )
                elif ioption == 'l':
                    if local_ylogscale:
                        local_ylogscale = False
                    else:
                        local_ylogscale = True
                elif ioption == 'p':
                    poldeg_residuals = readi('New polynomial degree',
                                             minval=0)
                elif ioption == 'o':
                    for i in range(len(ixpeaks)):
                        # identified lines
                        if wv_verified_all_peaks[i] > 0:
                            print(wv_verified_all_peaks[i],
                                  spectrum[ixpeaks[i]])
                elif ioption == 'x':
                    loop = False
                else:
                    print(wv_master)
                    expected_value = fxpeaks_wv[ioption -
                                                1] + polyres(fxpeaks_wv[ioption - 1])
                    print(">>> Current expected wavelength: ", expected_value)
                    delta_wv_max = ntimes_match_wv * cdelt1
                    close_value = match_wv_arrays(
                        wv_master,
                        np.array([expected_value]),
                        delta_wv_max=delta_wv_max)
                    newvalue = readf('New value (0 to delete line)',
                                     default=close_value[0])
                    wv_verified_all_peaks[ioption - 1] = newvalue
            else:
                loop = False
        else:
            loop = False

    # refined wavelength calibration coefficients
    if coeff_ini is not None:
        npoints_total = len(xresid)
        npoints_removed = sum(reject)
        npoints_used = npoints_total - npoints_removed
        if abs(debugplot) >= 10:
            print('>>> Npoints (total / used / removed)..:',
                  npoints_total, npoints_used, npoints_removed)
        if npoints_used < min_nlines_to_refine:
            print('Warning: number of lines insuficient to refine '
                  'wavelength calibration!')
            copc = 'n'
        else:
            if interactive:
                copc = readc('Refine wavelength calibration coefficients: '
                             '(y)es, (n)o', default='y', valid='yn')
            else:
                copc = 'y'
        if copc == 'y':
            coeff_refined = update_poly_wlcalib(
                coeff_ini=coeff_ini,
                coeff_residuals=polyres.coef,
                naxis1_ini=naxis1_ini,
                debugplot=0
            )
        else:
            coeff_refined = np.array(coeff_ini)
    else:
        coeff_refined = None

    if abs(debugplot) % 10 != 0:
        if coeff_refined is not None:
            for idum, fdum in \
                    enumerate(zip(coeff_ini, coeff_refined)):
                print(f">>> coef#{idum}:  ", end='')
                print(
                    f"{decimal.Decimal(fdum[0]):+.8E}  -->  {decimal.Decimal(fdum[1]):+.8E}")

    return coeff_refined


def update_poly_wlcalib(coeff_ini, coeff_residuals, naxis1_ini, debugplot):
    """Update wavelength calibration polynomial using the residuals fit.

    The idea is to repeat the original fit using the information
    previously computed with the function check_wlcalib_sp() in this
    module.

    Parameters
    ----------
    coeff_ini : array like (floats)
        Coefficients corresponding to the initial wavelength
        calibration.
    coeff_residuals: array like (floats)
        Coefficients corresponding to the fit performed by the
        function check_wlcalib_sp() in this module.
    naxis1_ini : int
        NAXIS1 in original spectrum employed to fit the initial
        wavelength calibration.
    debugplot : int
        Debugging level for messages and plots. For details see
        'numina.array.display.pause_debugplot.py'.

    Returns
    -------
    coeff_end : numpy array (floats)
        Updated coefficients.

    """

    # define initial wavelength calibration polynomial (use generic
    # code valid for lists of numpy.arrays)
    coeff = []
    for fdum in coeff_ini:
        coeff.append(fdum)
    poly_ini = np.polynomial.Polynomial(coeff)
    poldeg_wlcalib = len(coeff) - 1

    # return initial polynomial when there is no need to compute an
    # updated version
    if len(coeff_residuals) == 0:
        return poly_ini.coef
    else:
        if np.count_nonzero(poly_ini.coef) == 0:
            return poly_ini.coef

    # define polynomial corresponding to the residuals fit carried
    # out by check_wlcalib_sp()
    coeff = []
    for fdum in coeff_residuals:
        coeff.append(fdum)
    poly_residuals = np.polynomial.Polynomial(coeff)

    # define new points to be fitted
    xfit = np.zeros(naxis1_ini)
    yfit = np.zeros(naxis1_ini)
    for i in range(naxis1_ini):
        xfit[i] = float(i + 1)
        wv_tmp = poly_ini(xfit[i])
        yfit[i] = wv_tmp + poly_residuals(wv_tmp)

    # fit to get the updated polynomial
    if len(xfit) > poldeg_wlcalib:
        poldeg_effective = poldeg_wlcalib
    else:
        poldeg_effective = len(xfit) - 1
    poly_updated, ydum = polfit_residuals(
        x=xfit,
        y=yfit,
        deg=poldeg_effective,
        debugplot=debugplot
    )

    # return coefficients of updated polynomial
    return poly_updated.coef


def main(args=None):
    # parse command-line options
    parser = argparse.ArgumentParser()

    # positional parameters
    parser.add_argument("filename",
                        help="FITS image containing the spectra",
                        type=argparse.FileType('rb'))
    parser.add_argument("--scans", required=True,
                        help="Tuple ns1[,ns2] (from 1 to NAXIS2) to compute "
                             "median spectrum")
    parser.add_argument("--wv_master_file", required=True,
                        help="TXT file containing wavelengths",
                        type=argparse.FileType('rt'))

    # optional arguments
    parser.add_argument("--interactive",
                        help="Allows the user to modify de residuals fit",
                        action="store_true")
    parser.add_argument("--threshold",
                        help="Minimum signal in the line peaks (default=0)",
                        default=0, type=float)
    parser.add_argument("--nwinwidth_initial",
                        help="Window width where each peak must be found "
                             "(default=7)",
                        type=int, default=7)
    parser.add_argument("--nwinwidth_refined",
                        help="Window width where each peak must be refined "
                             "(default=5)",
                        type=int, default=5)
    parser.add_argument("--ntimes_match_wv",
                        help="Times CDELT1 to match measured and expected "
                             "wavelengths (default 2)",
                        default=2, type=float)
    parser.add_argument("--poldeg_residuals",
                        help="Polynomial degree for fit to residuals "
                             "(default 1)",
                        default=1, type=int)
    parser.add_argument("--times_sigma_reject",
                        help="Times the standard deviation to reject points "
                             "iteratively in the fit to residuals ("
                             "default=5)",
                        default=5, type=float)
    parser.add_argument("--use_r",
                        help="Perform additional statistical analysis with R",
                        action="store_true")
    parser.add_argument("--remove_null_borders",
                        help="Remove leading and trailing zeros in spectrum",
                        action="store_true")
    parser.add_argument("--geometry",
                        help="tuple x,y,dx,dy",
                        default="0,0,640,480")
    parser.add_argument("--debugplot",
                        help="integer indicating plotting/debugging" +
                             " (default=0)",
                        type=int, default=12,
                        choices=DEBUGPLOT_CODES)

    args = parser.parse_args(args=args)

    # scan range
    tmp_str = args.scans.split(",")
    if len(tmp_str) == 2:
        ns1 = int(tmp_str[0])
        ns2 = int(tmp_str[1])
    elif len(tmp_str) == 1:
        ns1 = int(tmp_str[0])
        ns2 = ns1
    else:
        raise ValueError("Invalid tuple for scan range")

    # geometry
    if args.geometry is None:
        geometry = None
    else:
        tmp_str = args.geometry.split(",")
        x_geom = int(tmp_str[0])
        y_geom = int(tmp_str[1])
        dx_geom = int(tmp_str[2])
        dy_geom = int(tmp_str[3])
        geometry = x_geom, y_geom, dx_geom, dy_geom

    # read FITS file
    with fits.open(args.filename) as hdulist:
        image2d_header = hdulist[0].header
        image2d = hdulist[0].data
    if image2d.ndim == 1:
        naxis1 = image2d.shape[0]
        naxis2 = 1
    elif image2d.ndim == 2:
        naxis2, naxis1 = image2d.shape
    else:
        raise ValueError("Unexpected image dimensions!")
    crpix1 = image2d_header['crpix1']
    crval1 = image2d_header['crval1']
    cdelt1 = image2d_header['cdelt1']
    print('* Input file:', args.filename.name)
    print('>>> NAXIS1:', naxis1)
    print('>>> NAXIS2:', naxis2)
    print('>>> CRPIX1:', crpix1)
    print('>>> CRVAL1:', crval1)
    print('>>> CDELT1:', cdelt1)

    if 1 <= ns1 <= ns2 <= naxis2:
        if ns1 == ns2 == 1 and image2d.ndim == 1:
            spmedian = np.copy(image2d[:])
        else:
            # extract spectrum
            spmedian = np.median(image2d[(ns1-1):ns2], axis=0)
    else:
        raise ValueError("Invalid scan numbers")

    # read list of expected arc lines
    master_table = np.genfromtxt(args.wv_master_file)
    wv_master = master_table[:, 0]
    if abs(args.debugplot) in (21, 22):
        print('wv_master:', wv_master)

    # define plot title
    title = 'fitsfile: ' + os.path.basename(args.filename.name) + \
            ' [' + str(ns1) + ',' + str(ns2) + ']\n' + \
            'wv_master: ' + os.path.basename(args.wv_master_file.name)

    # check the wavelength calibration
    check_wlcalib_sp(sp=spmedian,
                     crpix1=crpix1,
                     crval1=crval1,
                     cdelt1=cdelt1,
                     wv_master=wv_master,
                     interactive=args.interactive,
                     threshold=args.threshold,
                     nwinwidth_initial=args.nwinwidth_initial,
                     nwinwidth_refined=args.nwinwidth_refined,
                     ntimes_match_wv=args.ntimes_match_wv,
                     poldeg_residuals=args.poldeg_residuals,
                     times_sigma_reject=args.times_sigma_reject,
                     use_r=args.use_r,
                     title=title,
                     remove_null_borders=args.remove_null_borders,
                     geometry=geometry,
                     debugplot=args.debugplot)


if __name__ == "__main__":

    main()



def test_ini1(pattri_ins):
    pattri_ins.mode = "Mode_B"
    assert pattri_ins.get_property("mode") == "Mode_B"
    assert pattri_ins.get_property("pos") == 0
    assert pattri_ins.get_property("filter") == "U"
    assert pattri_ins.get_property("key1") == 0
    assert pattri_ins.get_property("key11") == 23
    assert pattri_ins.get_property("key12") == 12
    assert pattri_ins.get_property("key13") == 8
    assert pattri_ins.get_property("key14") == 0
    assert pattri_ins.get_property("key2") == "A"
    assert pattri_ins.get_property("key3") == "B"
    assert pattri_ins.get_property("tag0") == "tag0_A"
    assert pattri_ins.get_property("tag1") == "tag1_X"

    assert pattri_ins.get_property("PSU.csupos") == 444
    assert pattri_ins.get_property("PSU.box0") == 1
    # FIXME: is different to the value of filter
    assert pattri_ins.get_property("PSU.box1") == 134
    assert pattri_ins.get_property("PSU.tag0") == "tag0_A"
    assert pattri_ins.get_property("PSU.tag1") == "tag1_X"
    assert pattri_ins.get_property("PSU.mode") == "Mode_B"
    # This should be
    assert pattri_ins.get_property("PSU.spaces") == []
    assert pattri_ins.get_property("PSU.other") == [-1]


def test_ini2(pattri_ins):
    pattri_ins.mode = "Mode_B"
    pattri_ins.filter = "I"
    assert pattri_ins.get_property("PSU.box1") == -300


def test_ini3(pattri_ins):
    pattri_ins.mode = "Mode_B"
    pattri_ins.filter = "I"
    assert pattri_ins.get_property("PSU.spaces") == []


def test_depends(pattri_ins):
    assert pattri_ins.depends_on() == {"filter", "insmode", "key2", "tag0", "tag1"}
    assert pattri_ins.get_device("wheel").depends_on() == {"filter"}
    assert pattri_ins.get_device("PSU").depends_on() == {
        "filter",
        "insmode",
        "tag0",
        "tag1",
    }
    assert pattri_ins.get_device("robot").depends_on() == set()
    assert pattri_ins.get_device("detector").depends_on() == set()



# Copyright 2008-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import datetime
import uuid

import numpy

from numina.types.qc import QC

from .datatype import DataType
from .product import DataProductMixin


class DataProductType(DataProductMixin, DataType):
    def __init__(self, ptype, default=None):
        super(DataProductType, self).__init__(ptype, default=default)


class LinesCatalog(DataProductType):
    def __init__(self):
        super(LinesCatalog, self).__init__(ptype=numpy.ndarray)

    def __numina_load__(self, obj):
        with open(obj, 'r') as fd:
            linecat = numpy.loadtxt(fd)
        return linecat

    def extract_db_info(self, obj, keys):
        """Extract metadata from serialized file"""

        objl = self.convert(obj)

        result = super(LinesCatalog, self).extract_db_info(objl, keys)

        result['tags'] = {}
        result['type'] = 'LinesCatalog'
        result['uuid'] = str(uuid.uuid1())
        result['observation_date'] = datetime.datetime.now(
            datetime.timezone.utc)
        result['quality_control'] = QC.GOOD
        result['origin'] = {}

        return result



#
# Copyright 2017-2020 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#


class DepLink(object):
    """
    Dependency Link
    """

    def __init__(self, node, weight=1):
        self.node = node
        self.weight = weight


class DepNode(object):
    """Dependency Node"""

    def __init__(self, name, links=None):
        self.name = name
        self.links = [] if links is None else links

    def showtree(self):
        visited = []
        # print node.name
        visited.append(self.name)
        visit_node(self, visited, level=1)


def visit_node(node, visited, level=1):
    """
    Visit nodes in a dependency tree
    Parameters
    ----------
    node
    visited
    level

    Returns
    -------

    """
    # filler = 3 * level * "-"
    for nlk in node.links:
        # print filler, "> link,",level,", w=",l.weight, " dest=", l.node.name
        # print filler, ">", l.node.name
        if nlk.node.name in visited:
            continue
        else:
            visited.append(nlk.node.name)
        visit_node(nlk.node, visited, level+1)



#
# Copyright 2015-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import pytest

import numina.types.qc as qct
import numina.types.datatype as df
import numina.core.dataholders as dh

from ..recipeinout import RecipeResult, RecipeInput


class RRTest(RecipeResult):
    param1 = dh.Result(df.NullType, 'something1')
    param2 = dh.Result(df.NullType, 'something2')

    def mayfun(self):
        pass


def test_test1():

    m = RecipeResult()
    assert hasattr(m, 'qc')


def test_test2():
    m = RecipeResult()
    assert m.qc == qct.QC.UNKNOWN


def test_test3():
    m = RRTest(param1=None, param2=None)
    assert m.qc == qct.QC.UNKNOWN

    m = RRTest(param1=None, param2=None, qc=qct.QC.GOOD)
    assert m.qc == qct.QC.GOOD


@pytest.mark.parametrize("qc", [
    qct.QC.GOOD,
    qct.QC.UNKNOWN,
    qct.QC.BAD,
    qct.QC.PARTIAL,
])
def test_test4(qc):

    m = RRTest(param1=None, param2=None, qc=qc)
    assert m.qc == qc


@pytest.mark.parametrize("qc", [
    qct.QC.GOOD,
    qct.QC.UNKNOWN,
    qct.QC.BAD,
    qct.QC.PARTIAL,
])
def test_store_to(qc):
    result = RRTest(param1=None, param2=None, qc=qc)

    expected_result = {
        'values': {'param1': None, 'param2': None},
        'qc': qc.name,  # 'uuid': '00000000-0000-0000-0000-000000000000'
    }

    class Storage(object):
        def __init__(self):
            self.runinfo = {}

    where = Storage()

    saved_result = result.store_to(where)

    assert 'uuid' in saved_result

    # Do not check uuid field
    del saved_result['uuid']
    assert saved_result == expected_result


def test_capture_conversion_error():

    class RRTest1(RecipeInput):
        param1 = dh.Parameter(value=[100], description='some1')

    with pytest.raises(ValueError):
        RRTest1(param1=100)



import numina.core.pipeline

import pytest

from ..drpbase import DrpBase


def test_drpbase():
    drpbase = DrpBase()

    with pytest.raises(KeyError):
        drpbase.query_by_name('TEST1')

    assert drpbase.query_all() == {}


def test_invalid_instrument1():

    class Something(object):
        pass

    drpbase = DrpBase()
    assert drpbase.instrumentdrp_check(Something(), 'TEST1') is False


def test_invalid_instrument1_warning():

    with pytest.warns(RuntimeWarning):
        test_invalid_instrument1()


def test_invalid_instrument2():
    insdrp = numina.core.pipeline.InstrumentDRP('MYNAME', {}, {}, [], [])

    drpbase = DrpBase()
    res = drpbase.instrumentdrp_check(insdrp, 'TEST1')
    assert res is False


def test_invalid_instrument2_warning():

    insdrp = numina.core.pipeline.InstrumentDRP('MYNAME', {}, {}, [], [])

    drpbase = DrpBase()
    with pytest.warns(RuntimeWarning):
        drpbase.instrumentdrp_check(insdrp, 'TEST1')


def test_valid_instrument():
    insdrp = numina.core.pipeline.InstrumentDRP('TEST1', {}, {}, [], [])

    drpbase = DrpBase()
    res = drpbase.instrumentdrp_check(insdrp, 'TEST1')
    assert res



# Copyright 2008-2021 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import warnings

from astropy.io import fits

from numina.ext.gtc import DF
from numina.frame.schema import Schema
from .dataframe import DataFrame
from .datatype import DataType

_base_schema = {
    'keywords': {
        'INSTRUME': {'valid': True},
        'READMODE': {'valid': True},
        'EXPTIME': {'value': float},
        'NUMINAID': {'value': int}
    }
}


class DataFrameType(DataType):
    """A type of DataFrame."""

    db_info_keys = [
        'instrument',
        'object',
        'observation_date',
        'uuid',
        'type',
        'mode',
        'exptime',
        'darktime',
        # 'insconf',
        # 'blckuuid',
        'quality_control'
    ]

    tags_keys = []

    def __init__(self, datamodel=None):
        super(DataFrameType, self).__init__(DataFrame, datamodel=datamodel)
        self.headerschema = Schema(_base_schema)

        self.add_dialect_info('gtc', DF.TYPE_FRAME)

    def convert(self, obj):
        """Convert"""
        # We accept None representing No Image
        if obj is None:
            return None
        elif isinstance(obj, str):
            return DataFrame(filename=obj)
        elif isinstance(obj, DataFrame):
            return obj
        elif isinstance(obj, fits.HDUList):
            return DataFrame(frame=obj)
        elif isinstance(obj, fits.PrimaryHDU):
            return DataFrame(frame=fits.HDUList([obj]))
        else:
            msg = f'object of type {obj!r} cannot be converted to DataFrame'
            raise TypeError(msg)

    def validate(self, value):
        """validate"""
        # value must be None or convertible to HDUList
        # obj can be None or a DataFrame
        if value is None:
            return True
        else:
            if isinstance(value, fits.HDUList):
                hdulist = value
            else:
                hdulist = value.open()

            self.validate_hdulist(hdulist)

    def validate_hdulist(self, hdulist):
        pass

    def _datatype_dump(self, obj, where):
        return dump_dataframe(obj, where)

    def _datatype_load(self, obj):
        if obj is None:
            return None
        else:
            return DataFrame(filename=obj)

    def extract_db_info(self, obj, keys):
        """Extract tags from serialized file"""

        objl = self.convert(obj)

        result = super(DataFrameType, self).extract_db_info(objl, keys)
        ext = self.datamodel.extractor_map['fits']
        if objl:
            with objl.open() as hdulist:
                for field in keys:
                    result[field] = ext.extract(field, hdulist)

                tags = result['tags']
                # for field in self.tags_keys:
                for field in self.names_t:
                    tags[field] = ext.extract(field, hdulist)

                return result
        else:
            return result


def get_filename_dataframe(obj, where):
    # save fits file
    if obj is None or obj.frame is None:
        return None
    elif obj.filename:
        return obj.filename
    else:
        return get_filename_hdulist(obj, where)


def get_filename_hdulist(hdul, hint):
    ext = '.fits'

    if isinstance(hint, str):
        filename = hint + ext
    elif callable(hint):
        filename = hint(hdul)
    else:
        raise ValueError('hint is neither string nor callable')
    return filename


def dump_dataframe(obj, where):
    # save fits file

    fname = get_filename_dataframe(obj, where)

    if fname:
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            obj.frame.writeto(fname, overwrite=True, output_verify='warn')
    return fname



#
# Copyright 2008-2024 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

from astropy.io.fits import HDUList, Header


class QueryAttribute:
    def __init__(self, name, tipo, description=""):
        self.name = name
        self.type = tipo
        self.description = description


class KeyDefinition:
    def __init__(
        self, key: str, ext: int | str | None = None, default=None, convert=None
    ):
        self.key = key
        self.ext = 0 if ext is None else ext
        self.default = default
        # convert must be callable or None
        self.convert = convert
        # if selector is None:
        #    def selector(x): return x  # noqa: E731

    def _get_header(self, head: HDUList | Header) -> Header:
        match head:
            case HDUList():
                return head[self.ext].header
            case Header():
                return head
            case _:
                raise ValueError("head is not HDUList nor Header")

    def get_value(self, head: HDUList | Header):
        hdr = self._get_header(head)
        value = hdr.get(self.key, self.default)
        if self.convert:
            return self.convert(value)
        return value

    def set_value(self, head: HDUList | Header, value=None):
        hdr = self._get_header(head)

        if value is None:
            true_value = self.default
        else:
            if self.convert:
                true_value = self.convert(value)
            else:
                true_value = value

        hdr[self.key] = true_value

    def __call__(self, head: HDUList | Header):
        return self.get_value(head)


class FITSKeyExtractor:
    """Extract values from FITS images"""

    def __init__(self, values):
        self.map = {}
        for key, entry in values.items():
            if isinstance(entry, KeyDefinition):
                newval = entry
            elif isinstance(entry, tuple):
                if len(entry) == 3:
                    keyname = entry[0]
                    hduname = entry[1]
                    convert = entry[2]
                    default = None
                elif len(entry) == 2:
                    keyname = entry[0]
                    default = entry[1]
                    hduname = 0
                    convert = None
                else:
                    raise ValueError(
                        f"a tuple in FITSKeyExtractor must have 2-3 fields, has {len(entry)} instead"
                    )

                newval = KeyDefinition(
                    keyname, ext=hduname, convert=convert, default=default
                )
            elif isinstance(entry, str):
                newval = KeyDefinition(entry)
            else:
                newval = entry

            self.map[key] = newval

    def extract(self, value, hdulist):
        extractor = self.map[value]
        return extractor(hdulist)


def extract(header, conf, path, key_def):
    """Extracts values from a config dictionary and stores them in a header."""
    m = conf
    try:
        for part in path:
            m = m[part]
        key_def.set_value(header, m)
    except KeyError:
        # Keyword missing
        key_def.set_value(header)
    return header



#
# Copyright 2017 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Convert strings to functions in data load """

try:
    from functools import singledispatch
except ImportError:
    from pkgutil import simplegeneric as singledispatch

import numpy.polynomial.polynomial as nppol
from scipy.interpolate import UnivariateSpline


def json_deserial_func_u(args):
    """
    Rebuild UnivariateSpline

    Parameters
    ----------
    args

    Returns
    -------
    UnivariateSpline
    """
    # rebuild UnivariateSpline
    return UnivariateSpline._from_tck(args)


def json_deserial_func_p(args):
    """
    Rebuild Polynomial
    Parameters
    ----------
    args

    Returns
    -------
    nppol.Polynomial
    """
    if args:
        value = nppol.Polynomial(args)
    else:
        value = nppol.Polynomial([0.0])
    return value


_json_deserial_func_map = {
    'spline1d': json_deserial_func_u,
    'polynomial': json_deserial_func_p
}


def convert_function(node):
    tipo = node['function']
    args = node['params']

    json_deserial_func = _json_deserial_func_map[tipo]
    value = json_deserial_func(args)

    return value


@singledispatch
def json_serial_function(_):
    return {}


@json_serial_function.register(UnivariateSpline)
def json_serial_u(val):
    serial = {}
    # This is generic and should be somewhere else
    serial['function'] = 'spline1d'
    serial['params'] = val._eval_args
    return serial


@json_serial_function.register(nppol.Polynomial)
def json_serial_p(val):
    serial = {}
    # This is generic and should be somewhere else
    serial['function'] = 'polynomial1d'
    serial['params'] = val.coef
    return serial



import pytest

from ..shutter import Shutter


@pytest.fixture
def shutter_dev():
    wheel = Shutter()
    return wheel


def test_shutter1(shutter_dev):
    curr = shutter_dev.current()
    assert shutter_dev.pos() == 1
    assert curr.name == "OPEN"


def test_shutter_open(shutter_dev):
    shutter_dev.open()
    curr = shutter_dev.current()
    assert shutter_dev.pos() == 1
    assert curr.name == "OPEN"


def test_shutter_stop(shutter_dev):
    shutter_dev.close()
    curr = shutter_dev.current()
    assert shutter_dev.pos() == 0
    assert curr.name == "STOP"



#
# Copyright 2024 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Auxiliary function to display couloured text."""


def ctext(s=None,
          fg=None,
          bg=None,
          under=False,
          rev=False,
          faint=False,
          bold=False):
    """Return coloured string using ANSI Escape Sequences

    See ANSI Escape values in:
    https://gist.github.com/fnky/458719343aabd01cfb17a3a4f7296797

    Parameters
    ----------
    s : object
        Object to be displayed
    fg : str
        Foreground colour.
    bg : str
        Background colour.
    under : bool
        If True, add underline style.
    rev : bool
        If True, add reverse style.
    faint : bool
        If True, add dim/faint style.
    bold : bool
        If True, add bold style

    Returns
    -------
    result : str
        String containing the ANSI code required to display the
        coloured text.

    """

    colour = {
        'black': 0,
        'red': 1,
        'green': 2,
        'yellow': 3,
        'blue': 4,
        'magenta': 5,
        'cyan': 6,
        'white': 7,
        'default': 9,
        'reset': 0
    }

    if s is None:
        print(f'Available colours: {list(colour.keys())}')
        return

    # foreground
    if fg is not None:
        fg = fg.lower()
        if fg not in colour:
            raise ValueError(f'Unexpected foreground colour: {fg}')

    # background
    if bg is not None:
        bg = bg.lower()
        if bg not in colour:
            raise ValueError(f'Unexpected background colour: {bg}')

    style_list = []
    if under:
        style_list.append('\x1B[4m')
    if rev:
        style_list.append('\x1B[7m')
    if faint:
        style_list.append('\x1B[2m')
    if bold:
        style_list.append('\x1B[1m')

    if fg is not None:
        style_list.append(f'\x1B[3{colour[fg]}m')

    if bg is not None:
        style_list.append(f'\x1B[4{colour[bg]}m')

    final_style = ''.join(style_list)

    result = f'{final_style}{s}\x1B[0m'

    return result



#
# Copyright 2024-2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import matplotlib.pyplot as plt
import numpy as np


def fapply_atmosphere_transmission(simulated_wave, wave_transmission, curve_transmission, rng,
                                   plots=False, verbose=False):
    """Apply atmosphere transmission.

    The input wavelength of each photon is converted into -1
    if the photon is absorbed. These photons are discarded later
    when the code removes those outside [wmin, vmax].

    Parameters
    ----------
    simulated_wave : `~astropy.units.Quantity`
        Array containing the simulated wavelengths. If the photon is
        absorbed, the wavelength is changed to -1. Note that this
        input array is also the output of this function.
    wave_transmission : `~astropy.units.Quantity`
        Wavelength column of the tabulated transmission curve.
    curve_transmission : `~astropy.units.Quantity`
        Transmission values for the wavelengths given in
        'wave_transmission'.
    rng : `~numpy.random._generator.Generator`
        Random number generator.
    plots : bool
        If True, plot input and output results.
    verbose : bool
        If True, display additional information.

    """

    wave_unit = simulated_wave.unit

    # compute transmission at the wavelengths of the simulated photons
    transmission_values = np.interp(
        x=simulated_wave.value,
        xp=wave_transmission.to(wave_unit).value,
        fp=curve_transmission
    )

    if plots:
        fig, ax = plt.subplots()
        ax.plot(wave_transmission.to(wave_unit), curve_transmission, '-', label='SKYCALC curve')
        ax.plot(simulated_wave, transmission_values, ',', alpha=0.5, label='interpolated values')
        ax.set_xlabel(f'Wavelength ({wave_unit})')
        ax.set_ylabel('Transmission fraction')
        ax.legend(loc=3)  # loc="best" can be slow with large amounts of data
        plt.tight_layout()
        plt.show()

    # generate random values in the interval [0, 1] and discard photons whose
    # transmission value is lower than the random value
    nphotons = len(simulated_wave)
    survival_probability = rng.uniform(low=0, high=1, size=nphotons)
    iremove = np.argwhere(transmission_values < survival_probability)
    simulated_wave[iremove] = -1 * wave_unit

    if verbose:
        print('Applying atmosphere transmission:')
        print(f'- initial number of photons: {nphotons}')
        textwidth_nphotons_number = len(str(nphotons))
        percentage = np.round(100 * len(iremove) / nphotons, 2)
        print(f'- number of photons removed: {len(iremove): > {textwidth_nphotons_number}}  ({percentage}%)')



#
# Copyright 2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#
"""Diagnostic plot for cosmic ray mask computation."""
import sys

import matplotlib.pyplot as plt
from matplotlib.widgets import TextBox
import numpy as np
import teareduce as tea


def segregate_cr_flags(naxis1, naxis2, flag_only_la, flag_only_sb, flag_both,
                       enum_la_global, enum_mm_global, enum_both_global,
                       within_xy_diagram):
    """Segregate the cosmic ray flags into three categories:
    - detected only by the lacosmic method
    - detected only by the mmcosmic method
    - detected by both methods

    Parameters
    ----------
    naxis1 : int
        The size of the first dimension of the input arrays.
    naxis2 : int
        The size of the second dimension of the input arrays.
    flag_only_la : 1D numpy array
        A boolean array indicating which pixels are detected
        only by the lacosmic method.
    flag_only_sb : 1D numpy array
        A boolean array indicating which pixels are detected
        only by the mmcosmic method.
    flag_both : 1D numpy array
        A boolean array indicating which pixels are detected
        by both methods.
    enum_la_global : 1D numpy array
        An integer array with the enumeration of the pixels
        detected only by the lacosmic method.
    enum_mm_global : 1D numpy array
        An integer array with the enumeration of the pixels
        detected only by the mmcosmic method.
    enum_both_global : 1D numpy array
        An integer array with the enumeration of the pixels
        detected by both methods.
    within_xy_diagram : 1D numpy array
        A boolean array indicating which pixels are within the XY diagram.

    Returns
    -------
    flag_only_la_within_xy : 1D numpy array
        A boolean array indicating which pixels are detected
        only by the lacosmic method within the XY diagram.
    flag_only_mm_within_xy : 1D numpy array
        A boolean array indicating which pixels are detected
        only by the mmcosmic method within the XY diagram.
    flag_both_within_xy : 1D numpy array
        A boolean array indicating which pixels are detected
        by both methods within the XY diagram.
    (num_only_la, xcr_only_la, ycr_only_la) : tuple
        Number of pixels detected only by the lacosmic method,
        and their x and y coordinates (FITS convention; first pixel is (1, 1)).
        If no pixels are detected, xcr_only_la and ycr_only_la are None.
    (num_only_sb, xcr_only_sb, ycr_only_sb) : tuple
        Number of pixels detected only by the mmcosmic method,
        and their x and y coordinates (FITS convention; first pixel is (1, 1)).
        If no pixels are detected, xcr_only_sb and ycr_only_sb are None.
    (num_both, xcr_both, ycr_both) : tuple
        Number of pixels detected by both methods,
        and their x and y coordinates (FITS convention; first pixel is (1, 1)).
        If no pixels are detected, xcr_both and ycr_both are None.
    """

    # Segregate the cosmic rays within the XY diagnostic diagram
    flag_only_la_within_xy = flag_only_la & within_xy_diagram
    flag_only_mm_within_xy = flag_only_sb & within_xy_diagram
    flag_both_within_xy = flag_both & within_xy_diagram

    num_only_la_within_xy = np.sum(flag_only_la_within_xy)
    if num_only_la_within_xy > 0:
        pixels_detected = np.argwhere(flag_only_la_within_xy.reshape(naxis2, naxis1))
        xcr_only_la_within_xy = pixels_detected[:, 1] + 1  # FITS convention: first pixel is (1, 1)
        ycr_only_la_within_xy = pixels_detected[:, 0] + 1  # FITS convention: first pixel is (1, 1)
        ncr_only_la_within_xy = enum_la_global[flag_only_la_within_xy]
    else:
        xcr_only_la_within_xy, ycr_only_la_within_xy, ncr_only_la_within_xy = None, None, None

    num_only_mm_within_xy = np.sum(flag_only_mm_within_xy)
    if num_only_mm_within_xy > 0:
        pixels_detected = np.argwhere(flag_only_mm_within_xy.reshape(naxis2, naxis1))
        xcr_only_mm_within_xy = pixels_detected[:, 1] + 1  # FITS convention: first pixel is (1, 1)
        ycr_only_mm_within_xy = pixels_detected[:, 0] + 1  # FITS convention: first pixel is (1, 1)
        ncr_only_mm_within_xy = enum_mm_global[flag_only_mm_within_xy]
    else:
        xcr_only_mm_within_xy, ycr_only_mm_within_xy, ncr_only_mm_within_xy = None, None, None

    num_both_within_xy = np.sum(flag_both_within_xy)
    if num_both_within_xy > 0:
        pixels_detected = np.argwhere(flag_both_within_xy.reshape(naxis2, naxis1))
        xcr_both_within_xy = pixels_detected[:, 1] + 1  # FITS convention: first pixel is (1, 1)
        ycr_both_within_xy = pixels_detected[:, 0] + 1  # FITS convention: first pixel is (1, 1)
        ncr_both_within_xy = enum_both_global[flag_both_within_xy]
    else:
        xcr_both_within_xy, ycr_both_within_xy, ncr_both_within_xy = None, None, None

    return \
        flag_only_la_within_xy, flag_only_mm_within_xy, flag_both_within_xy, \
        (num_only_la_within_xy, xcr_only_la_within_xy, ycr_only_la_within_xy, ncr_only_la_within_xy), \
        (num_only_mm_within_xy, xcr_only_mm_within_xy, ycr_only_mm_within_xy, ncr_only_mm_within_xy), \
        (num_both_within_xy, xcr_both_within_xy, ycr_both_within_xy, ncr_both_within_xy)


def update_marks(naxis1, naxis2, flag_only_la, flag_only_sb, flag_both,
                 enum_la_global, enum_mm_global, enum_both_global,
                 xplot, yplot,
                 ax1, ax2, ax3, display_ncr=True):
    """Update the marks in the diagnostic plot.
    If ax2 and ax3 are None, only the segregation of the cosmic rays
    within the XY diagram is performed and the information of the
    suspected pixels is printed in the terminal.
    """
    if flag_only_la.shape != (naxis2 * naxis1,):
        raise ValueError(f"{flag_only_la.shape=} must have shape (naxis2*naxis1,)={naxis1*naxis2}.")
    if flag_only_la.shape != flag_only_sb.shape:
        raise ValueError(f"{flag_only_sb.shape=} must have shape (naxis2*naxis1,)={naxis1*naxis2}.")
    if flag_only_la.shape != flag_both.shape:
        raise ValueError(f"{flag_both.shape=} must have shape (naxis2*naxis1,)={naxis1*naxis2}.")
    if flag_only_la.shape != xplot.shape or flag_only_la.shape != yplot.shape:
        raise ValueError(f"{xplot.shape=} and {yplot.shape=} must have {flag_only_la.shape=}.")

    xlim = ax1.get_xlim()
    ylim = ax1.get_ylim()
    within_xy_diagram = (xlim[0] <= xplot) & (xplot <= xlim[1]) & (ylim[0] <= yplot) & (yplot <= ylim[1])

    flag_only_la_within_xy, flag_only_mm_within_xy, flag_both_within_xy, tuple_la, tuple_sb, tuple_both = \
        segregate_cr_flags(naxis1, naxis2, flag_only_la, flag_only_sb, flag_both,
                           enum_la_global, enum_mm_global, enum_both_global,
                           within_xy_diagram)
    num_only_la_within_xy, xcr_only_la_within_xy, ycr_only_la_within_xy, ncr_only_la_within_xy = tuple_la
    num_only_mm_within_xy, xcr_only_mm_within_xy, ycr_only_mm_within_xy, ncr_only_mm_within_xy = tuple_sb
    num_both_within_xy, xcr_both_within_xy, ycr_both_within_xy, ncr_both_within_xy = tuple_both

    if ax2 is None and ax3 is None:
        ax_list = [None]
    else:
        ax_list = [ax2, ax3]
    for ax in ax_list:
        for num, method, xcr, ycr, ncr, flag_only, color, marker in zip(
                [num_only_la_within_xy, num_only_mm_within_xy, num_both_within_xy],
                ['lacosmic', 'mmcosmic', 'both'],
                [xcr_only_la_within_xy, xcr_only_mm_within_xy, xcr_both_within_xy],
                [ycr_only_la_within_xy, ycr_only_mm_within_xy, ycr_both_within_xy],
                [ncr_only_la_within_xy, ncr_only_mm_within_xy, ncr_both_within_xy],
                [flag_only_la_within_xy, flag_only_mm_within_xy, flag_both_within_xy],
                ['r', 'b', 'm'],
                ['x', '+', 'o']):
            if num > 0:
                if ax is None:
                    print("-" * 78)
                    print(f"{num} cosmic rays detected with method {method}.")
                    for ix, iy, ncr in zip(xcr, ycr, ncr):
                        print(f"  Pixel (x, y) = ({ix}, {iy}), number {ncr}")
                elif ax == ax2:
                    for ix, iy, ncr in zip(xplot[flag_only], yplot[flag_only], ncr):
                        ax.text(ix, iy, str(ncr), color=color, fontsize=8, clip_on=True,
                                ha='center', va='center', bbox=dict(facecolor='white', alpha=0.5))
                elif ax == ax3:
                    if display_ncr:
                        for ix, iy, ncr in zip(xcr, ycr, ncr):
                            ax.text(ix, iy, str(ncr), color=color, fontsize=8, clip_on=True,
                                    ha='center', va='center', bbox=dict(facecolor='white', alpha=0.5))
                    else:
                        if marker == 'o':
                            ax.scatter(xcr, ycr, edgecolors=color, marker=marker, facecolors='none')
                        else:
                            ax.scatter(xcr, ycr, c=color, marker=marker)


def diagnostic_plot(xplot, yplot, xplot_boundary, yplot_boundary, flag_la, flag_sb,
                    mm_threshold, ylabel, interactive, target2d, target2d_name,
                    min2d, mean2d, image3d,
                    _logger=None, png_filename=None):
    """Diagnostic plot for the mediancr function.
    """
    if png_filename is None:
        raise ValueError("png_filename must be provided for diagnostic plots.")

    # Set up relevant parameters
    naxis3, naxis2, naxis1 = image3d.shape
    if target2d.shape != (naxis2, naxis1):
        raise ValueError("target2d must have shape (naxis2, naxis1).")
    if min2d.shape != (naxis2, naxis1):
        raise ValueError("min2d must have shape (naxis2, naxis1).")
    if mean2d.shape != (naxis2, naxis1):
        raise ValueError("mean2d must have shape (naxis2, naxis1).")
    if flag_la.shape != (naxis2 * naxis1,):
        raise ValueError(f"{flag_la.shape=} must have shape (naxis2*naxis1,)={naxis1*naxis2}.")
    if flag_la.shape != flag_sb.shape:
        raise ValueError(f"{flag_sb.shape=} must have shape (naxis2*naxis1,)={naxis1*naxis2}.")
    if xplot.shape != (naxis2 * naxis1,):
        raise ValueError(f"{xplot.shape=} must have shape (naxis2*naxis1,)={naxis1*naxis2}.")
    if yplot.shape != (naxis2 * naxis1,):
        raise ValueError(f"{yplot.shape=} must have shape (naxis2*naxis1,)={naxis1*naxis2}.")

    display_ncr = False   # display the number of cosmic rays in the plot instead of symbols
    aspect_imshow = 'auto'  # 'equal' or 'auto'
    i_comparison_image = 0  # 0 for mean2d, 1, 2,... for image3d[comparison_image-1]

    if interactive:
        fig = plt.figure(figsize=(12, 8))
        x0_plot = 0.07
        y0_plot = 0.07
        width_plot = 0.4
        height_plot = 0.4
        vspace_plot = 0.09
        # top left, top right, bottom left, bottom right
        ax1 = fig.add_axes([x0_plot, y0_plot + height_plot + vspace_plot, width_plot, height_plot])
        ax2 = fig.add_axes([0.55, y0_plot + height_plot + vspace_plot, width_plot, height_plot],
                           sharex=ax1, sharey=ax1)
        ax3 = fig.add_axes([x0_plot, y0_plot, width_plot, height_plot])
        ax4 = fig.add_axes([0.55, y0_plot, width_plot, height_plot], sharex=ax3, sharey=ax3)
        dx_text = 0.07
        ax_vmin = fig.add_axes([x0_plot, y0_plot + height_plot + 0.005, dx_text, 0.03])
        ax_vmax = fig.add_axes([x0_plot + width_plot - dx_text, y0_plot + height_plot + 0.005, dx_text, 0.03])
    else:
        fig, axarr = plt.subplots(nrows=2, ncols=2, figsize=(12, 8))
        ax1, ax2, ax3, ax4 = axarr.flatten()
        ax_vmin, ax_vmax = None, None

    # Segregate the cosmic rays detected by the different methods
    flag_only_la = flag_la & ~flag_sb
    flag_only_sb = flag_sb & ~flag_la
    flag_both = flag_la & flag_sb
    num_only_la = np.sum(flag_only_la)
    num_only_sb = np.sum(flag_only_sb)
    num_both = np.sum(flag_both)
    num_total = num_only_la + num_only_sb + num_both

    # Enumerate the cosmic rays detected by the different methods
    enum_la_global = np.zeros_like(flag_la, dtype=int)
    enum_la_global[flag_only_la] = np.arange(1, np.sum(flag_only_la) + 1, dtype=int)
    enum_mm_global = np.zeros_like(flag_sb, dtype=int)
    enum_mm_global[flag_only_sb] = np.arange(1, np.sum(flag_only_sb) + 1, dtype=int)
    enum_both_global = np.zeros_like(flag_la, dtype=int)
    enum_both_global[flag_both] = np.arange(1, np.sum(flag_both) + 1, dtype=int)

    ax1.plot(xplot, yplot, 'C0,', label='Non-suspected pixels')
    ax1.scatter(xplot[flag_only_la], yplot[flag_only_la],
                c='r', marker='x', label=f'Suspected pixels: {num_only_la} (lacosmic)')
    ax1.scatter(xplot[flag_only_sb], yplot[flag_only_sb],
                c='b', marker='+', label=f'Suspected pixels: {num_only_sb} (mmcosmic)')
    ax1.scatter(xplot[flag_both], yplot[flag_both],
                edgecolor='m', marker='o', facecolors='none', label=f'Suspected pixels: {num_both} (both methods)')
    if xplot_boundary is not None and yplot_boundary is not None:
        ax1.plot(xplot_boundary, yplot_boundary, 'C1-', label='Detection boundary')
    if mm_threshold is not None:
        ax1.axhline(mm_threshold, color='gray', linestyle=':', label=f'mm_threshold ({mm_threshold:.2f})')
    ax1.set_xlabel(r'min2d $-$ bias')  # the bias was subtracted from the input arrays
    ax1.set_ylabel(ylabel)
    ax1.set_title('Median-Mean Diagnostic Diagram')
    ax1.legend(loc='upper right', fontsize=8, title=f'Total: {num_total} suspected pixels')

    ax2.set_xlim(ax1.get_xlim())
    ax2.set_ylim(ax1.get_ylim())
    ax2.set_xlabel(ax1.get_xlabel())
    ax2.set_ylabel(ax1.get_ylabel())
    ax2.set_title(ax1.get_title())

    vmin, vmax = tea.zscale(target2d)
    ax3_title = target2d_name
    img_ax3, _, _ = tea.imshow(fig, ax3, target2d, ds9mode=True, aspect=aspect_imshow, vmin=vmin, vmax=vmax,
                               title=ax3_title, cmap='viridis', colorbar=False)
    if i_comparison_image == 0:
        comparison_image = mean2d
        ax4_title = 'mean2d'
    else:
        comparison_image = image3d[i_comparison_image - 1]
        ax4_title = f'single exposure #{i_comparison_image}]'
    img_ax4, _, _ = tea.imshow(fig, ax4, comparison_image, ds9mode=True, aspect=aspect_imshow, vmin=vmin, vmax=vmax,
                               title=ax4_title, cmap='viridis', colorbar=False)
    ax3.set_title(ax3_title)
    ax4.set_title(ax4_title)
    update_marks(naxis1, naxis2, flag_only_la, flag_only_sb, flag_both,
                 enum_la_global, enum_mm_global, enum_both_global,
                 xplot, yplot,
                 ax1, ax2, ax3, display_ncr)

    for ax, label, color in zip([ax1, ax2, ax3, ax4],
                                ['(a)', '(b)', '(c)', '(d)'],
                                ['k', 'k', 'w', 'w']):
        ax.text(0.97, 0.03, label, transform=ax.transAxes, color=color,
                fontsize=15, fontweight='bold', va='bottom', ha='right')

    updating = {'plot_limits': False}

    def sync_zoom_x(event_ax):
        if updating['plot_limits']:
            return
        try:
            updating['plot_limits'] = True
            if event_ax is ax1:
                xlim = ax1.get_xlim()
                ax2.set_xlim(xlim)
            elif event_ax is ax2:
                pass
            elif event_ax is ax3:
                pass
            elif event_ax is ax4:
                pass
        finally:
            updating['plot_limits'] = False

    def sync_zoom_y(event_ax):
        nonlocal img_ax3, img_ax4
        nonlocal display_ncr
        if updating['plot_limits']:
            return
        try:
            updating['plot_limits'] = True
            if event_ax is ax1:
                ylim = ax1.get_ylim()
                ax2.set_ylim(ylim)
                xlim = ax3.get_xlim()
                ylim = ax3.get_ylim()
                ax3.cla()
                img_ax3, _, _ = tea.imshow(fig, ax3, target2d, ds9mode=True, aspect=aspect_imshow,
                                           vmin=vmin, vmax=vmax,
                                           title=ax3_title, cmap='viridis', colorbar=False)
                ax3.set_xlim(xlim)
                ax3.set_ylim(ylim)
                xlim = ax4.get_xlim()
                ylim = ax4.get_ylim()
                ax4.cla()
                img_ax4, _, _ = tea.imshow(fig, ax4, comparison_image, ds9mode=True, aspect=aspect_imshow,
                                           vmin=vmin, vmax=vmax,
                                           title=ax4_title, cmap='viridis', colorbar=False)
                ax4.set_xlim(xlim)
                ax4.set_ylim(ylim)
                update_marks(naxis1, naxis2, flag_only_la, flag_only_sb, flag_both,
                             enum_la_global, enum_mm_global, enum_both_global,
                             xplot, yplot,
                             ax1, ax2, ax3, display_ncr)
                for ax, label, color in zip([ax3, ax4],
                                            ['(c)', '(d)'],
                                            ['w', 'w']):
                    ax.text(0.97, 0.03, label, transform=ax.transAxes, color=color,
                            fontsize=15, fontweight='bold', va='bottom', ha='right')

                ax2.figure.canvas.draw_idle()
                ax3.figure.canvas.draw_idle()
                ax4.figure.canvas.draw_idle()
            elif event_ax is ax2:
                pass
            elif event_ax is ax3:
                pass
            elif event_ax is ax4:
                pass
        finally:
            updating['plot_limits'] = False

    ax1.callbacks.connect('xlim_changed', sync_zoom_x)
    ax1.callbacks.connect('ylim_changed', sync_zoom_y)

    if not interactive:
        plt.tight_layout()
    if png_filename is not None:
        _logger.info(f"saving {png_filename}")
        plt.savefig(png_filename, dpi=150)
    if interactive:
        init_limits = {ax: (ax.get_xlim(), ax.get_ylim()) for ax in [ax1, ax2, ax3, ax4]}

        mouse_info = {'ax': None, 'x': None, 'y': None}

        def on_mouse_move(event):
            if event.inaxes:
                mouse_info['ax'] = event.inaxes
                mouse_info['x'] = event.xdata
                mouse_info['y'] = event.ydata

        fig.canvas.mpl_connect('motion_notify_event', on_mouse_move)

        def on_key(event):
            nonlocal vmin, vmax
            nonlocal img_ax3, img_ax4
            nonlocal display_ncr
            nonlocal aspect_imshow
            nonlocal i_comparison_image, comparison_image
            ax_mouse = mouse_info['ax']
            x_mouse = mouse_info['x']
            y_mouse = mouse_info['y']

            if event.key in ("h", "H", "r", "R"):
                for ax in [ax1, ax2, ax3, ax4]:
                    init_xlim, init_ylim = init_limits[ax]
                    ax.set_xlim(init_xlim)
                    ax.set_ylim(init_ylim)
                if i_comparison_image == 0:
                    comparison_image = mean2d
                    ax4_title = 'mean2d'
                else:
                    comparison_image = image3d[i_comparison_image - 1]
                    ax4_title = f'single exposure #{i_comparison_image}'
                ax4.set_title(ax4_title)
            elif event.key == '?':
                _logger.info("-" * 79)
                _logger.info("Keyboard shortcuts:")
                _logger.info("'h' or 'r': reset zoom to initial limits")
                _logger.info("'p': pan mode")
                _logger.info("'o': zoom to rectangle")
                _logger.info("'f': toggle full screen mode")
                _logger.info("'s': save the figure to a PNG file")
                _logger.info("." * 79)
                _logger.info("'?': show this help message")
                _logger.info("'i': print pixel info at mouse position (ax3 only)")
                _logger.info("'&': print CR pixels within the zoomed region (ax3 only)")
                _logger.info("'n': toggle display of number of cosmic rays (ax3 only)")
                _logger.info("'a': toggle imshow aspect='equal' / aspect='auto' (ax3 and ax4 only)")
                _logger.info("'t': toggle mean2d -> individual exposures in ax4")
                _logger.info("'0': switch to mean2d in ax4")
                _logger.info("'1', '2', ...: switch to individual exposure #1, #2, ... in ax4")
                _logger.info("',': set vmin and vmax to min and max of the zoomed region (ax3 and ax4 only)")
                _logger.info("'/': set vmin and vmax using zscale of the zoomed region (ax3 and ax4 only)")
                _logger.info("'q': close the plot and continue the program execution")
                _logger.info("'x': halt the program execution")
                _logger.info("-" * 79)
            elif event.key == "i":
                if ax_mouse in [ax1, ax2]:
                    print(f'x_mouse = {x_mouse:.3f}, y_mouse = {y_mouse:.3f}')
                elif ax_mouse in [ax3]:
                    ix = int(round(x_mouse))
                    iy = int(round(y_mouse))
                    if 1 <= ix <= naxis1 and 1 <= iy <= naxis2:
                        print('-' * 79)
                        print(f'Pixel coordinates (FITS criterium): ix = {ix}, iy = {iy}')
                        print(f'target2d - min2d = {target2d[iy-1, ix-1] - min2d[iy-1, ix-1]:.3f}')
                        print(f'min2d - bias     = {min2d[iy-1, ix-1]:.3f}')
                        print('.' * 79)
                        for inum in range(image3d.shape[0]):
                            print(f'(image {inum+1} - bias) = {image3d[inum, iy-1, ix-1]:.3f}')
                        print('.' * 79)
                        for flag, crmethod in zip([flag_la, flag_sb, flag_both], ['lacosmic', 'mmcosmic']):
                            # Python convention: first pixel is (0, 0) but iy and ix are in FITS convention
                            # where the first pixel is (1, 1)
                            if flag.reshape((naxis2, naxis1))[iy-1, ix-1]:
                                print(f'Pixel found by {crmethod}')
                            else:
                                print(f'Pixel not found by {crmethod}')
            elif event.key == "&":
                if ax_mouse in [ax3]:
                    update_marks(naxis1, naxis2, flag_only_la, flag_only_sb, flag_both,
                                 enum_la_global, enum_mm_global, enum_both_global,
                                 xplot, yplot,
                                 ax1, None, None, None)
            elif event.key == 'n':
                if ax_mouse in [ax3]:
                    display_ncr = not display_ncr
                    sync_zoom_y(ax1)
            elif event.key == 'a':
                if aspect_imshow == 'equal':
                    aspect_imshow = 'auto'
                else:
                    aspect_imshow = 'equal'
                sync_zoom_y(ax1)
            elif event.key in ['t', '0', '1', '2', '3', '4', '5', '6', '7', '8', '9']:
                if ax_mouse == ax4:
                    i_comparison_image_previous = i_comparison_image
                    if event.key == 't':
                        i_comparison_image += 1
                        if i_comparison_image > naxis3:
                            i_comparison_image = 0
                    elif event.key == '0':
                        i_comparison_image = 0
                    else:
                        i_comparison_image = int(event.key)
                        if i_comparison_image > naxis3:
                            i_comparison_image = i_comparison_image_previous
                    if i_comparison_image != i_comparison_image_previous:
                        if i_comparison_image == 0:
                            comparison_image = mean2d
                            ax4_title = 'mean2d'
                        else:
                            comparison_image = image3d[i_comparison_image - 1]
                            ax4_title = f'single exposure #{i_comparison_image}'
                        print(f'Switching to {ax4_title} in ax4')
                        vmin, vmax = img_ax4.get_clim()
                        img_ax4.set_data(comparison_image)
                        img_ax4.set_clim(vmin=vmin, vmax=vmax)
                        ax4.set_title(ax4_title)
                        ax4.figure.canvas.draw_idle()
            elif event.key in [',', '/']:
                if ax_mouse in [ax3, ax4]:
                    # Determine the region in the image corresponding to the zoomed area
                    xmin, xmax = ax_mouse.get_xlim()
                    ymin, ymax = ax_mouse.get_ylim()
                    ixmin = int(round(xmin))
                    ixmax = int(round(xmax))
                    iymin = int(round(ymin))
                    iymax = int(round(ymax))
                    if ixmin > ixmax:
                        ixmin, ixmax = ixmax, ixmin
                    if iymin > iymax:
                        iymin, iymax = iymax, iymin
                    if ixmin < 1:
                        ixmin = 1
                    if ixmax > naxis1:
                        ixmax = naxis1
                    if iymin < 1:
                        iymin = 1
                    if iymax > naxis2:
                        iymax = naxis2
                    region2d = tea.SliceRegion2D(f'[{ixmin}:{ixmax},{iymin}:{iymax}]', mode='fits').python
                    if event.key == ',':
                        if ax_mouse == ax3:
                            vmin, vmax = np.min(target2d[region2d]), np.max(target2d[region2d])
                        elif ax_mouse == ax4:
                            vmin, vmax = np.min(comparison_image[region2d]), np.max(comparison_image[region2d])
                    elif event.key == '/':
                        if ax_mouse == ax3:
                            vmin, vmax = tea.zscale(target2d[region2d])
                        elif ax_mouse == ax4:
                            vmin, vmax = tea.zscale(comparison_image[region2d])
                    text_box_vmin.set_val(f'{int(np.round(vmin, 0))}')
                    text_box_vmax.set_val(f'{int(np.round(vmax, 0))}')

                    img_ax3.set_clim(vmin=vmin, vmax=vmax)
                    img_ax4.set_clim(vmin=vmin, vmax=vmax)

                    ax3.figure.canvas.draw_idle()
                    ax4.figure.canvas.draw_idle()
            elif event.key == 'x':
                _logger.info("Exiting program as per user request ('x' key pressed).")
                plt.close(fig)
                sys.exit(0)

        fig.canvas.mpl_connect("key_press_event", on_key)

        def submit_vmin(text):
            nonlocal vmin, vmax
            text = text.strip()
            if text:
                try:
                    vmin_ = float(text)
                    vmin = vmin_
                    img_ax3.set_clim(vmin=vmin, vmax=vmax)
                    img_ax4.set_clim(vmin=vmin, vmax=vmax)
                    ax3.figure.canvas.draw_idle()
                    ax4.figure.canvas.draw_idle()
                except ValueError:
                    print(f'Invalid input: {text}')

        def submit_vmax(text):
            nonlocal vmin, vmax
            text = text.strip()
            if text:
                try:
                    vmax_ = float(text)
                    vmax = vmax_
                    img_ax3.set_clim(vmin=vmin, vmax=vmax)
                    img_ax4.set_clim(vmin=vmin, vmax=vmax)
                    ax3.figure.canvas.draw_idle()
                    ax4.figure.canvas.draw_idle()
                except ValueError:
                    print(f'Invalid input: {text}')

        text_box_vmin = TextBox(ax_vmin, 'vmin:', initial=f'{int(np.round(vmin, 0))}', textalignment='right')
        text_box_vmin.on_submit(submit_vmin)
        text_box_vmax = TextBox(ax_vmax, 'vmax:', initial=f'{int(np.round(vmax, 0))}', textalignment='right')
        text_box_vmax.on_submit(submit_vmax)

        _logger.info("Entering interactive mode (press 'q' to close figure, 'x' to quit program)")
        # plt.tight_layout()
        plt.show()
    plt.close(fig)



#
# Copyright 2016-2018 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""
Validator decorator
"""

from functools import wraps

from numina.exceptions import ValidationError


def validate(method):
    """Decorate run method, inputs and outputs are validated"""

    @wraps(method)
    def mod_run(self, rinput):
        self.validate_input(rinput)
        #
        result = method(self, rinput)
        #
        self.validate_result(result)
        return result

    return mod_run


def only_positive(value):
    """Validation error is value is negative"""
    if value < 0:
        raise ValidationError("must be >= 0")
    return value


def as_list(callable):
    """Convert a scalar validator in a list validator"""
    @wraps(callable)
    def wrapper(value_iter):
        return [callable(value) for value in value_iter]

    return wrapper


def range_validator(minval=None, maxval=None):
    """Generates a function that validates that a number is within range

    Parameters
    ==========
    minval: numeric, optional:
        Values strictly lesser than `minval` are rejected
    maxval: numeric, optional:
        Values strictly greater than `maxval` are rejected

    Returns
    =======
    A function that returns values if are in the range and raises
    ValidationError is the values are outside the range

    """
    def checker_func(value):
        if minval is not None and value < minval:
            msg = f"must be >= {minval}"
            raise ValidationError(msg)
        if maxval is not None and value > maxval:
            msg = f"must be <= {maxval}"
            raise ValidationError(msg)
        return value

    return checker_func



#
# Copyright 2015-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Utility functions to handle image distortions."""

import numpy as np
from skimage import transform

from numina.array._clippix import _resample
from numina.array.display.pause_debugplot import pause_debugplot
from numina.array.display.ximplotxy import ximplotxy

NMAX_ORDER = 7


def compute_distortion(x_orig, y_orig, x_rect, y_rect, order, debugplot):
    """Compute image distortion transformation.

    This function computes the following 2D transformation:
    x_orig = sum[i=0:order]( sum[j=0:i]( a_ij * x_rect**(i - j) * y_rect**j ))
    y_orig = sum[i=0:order]( sum[j=0:i]( b_ij * x_rect**(i - j) * y_rect**j ))

    Parameters
    ----------
    x_orig : numpy array
        X coordinate of the reference points in the distorted image
    y_orig : numpy array
        Y coordinate of the reference points in the distorted image
    x_rect : numpy array
        X coordinate of the reference points in the rectified image
    y_rect : numpy array
        Y coordinate of the reference points in the rectified image
    order : int
        Order of the polynomial transformation
    debugplot : int
        Determine whether intermediate computations and/or plots
        are displayed. The valid codes are defined in
        numina.array.display.pause_debugplot.

    Returns
    -------

    aij : numpy array
        Coefficients a_ij of the 2D transformation.
    bij : numpy array
        Coefficients b_ij of the 2D transformation.

    """

    # protections
    npoints = len(x_orig)
    for xdum in [y_orig, x_rect, y_rect]:
        if len(xdum) != npoints:
            raise ValueError('Unexpected different number of points')
    if order < 1 or order > NMAX_ORDER:
        raise ValueError("Invalid order=" + str(order))

    # normalize ranges dividing by the maximum, so that the transformation
    # fit will be computed with data points with coordinates in the range [0,1]
    x_scale = 1.0 / np.concatenate((x_orig, x_rect)).max()
    y_scale = 1.0 / np.concatenate((y_orig, y_rect)).max()
    x_orig_scaled = x_orig * x_scale
    y_orig_scaled = y_orig * y_scale
    x_inter_scaled = x_rect * x_scale
    y_inter_scaled = y_rect * y_scale

    # solve 2 systems of equations with half number of unknowns each
    if order == 1:
        a_matrix = np.vstack([np.ones(npoints),
                              x_inter_scaled,
                              y_inter_scaled]
                             ).T
    elif order == 2:
        a_matrix = np.vstack([np.ones(npoints),
                              x_inter_scaled,
                              y_inter_scaled,
                              x_inter_scaled ** 2,
                              x_inter_scaled ** 1 * y_inter_scaled ** 1,
                              y_inter_scaled ** 2]
                             ).T
    elif order == 3:
        a_matrix = np.vstack([np.ones(npoints),
                              x_inter_scaled,
                              y_inter_scaled,
                              x_inter_scaled ** 2,
                              x_inter_scaled ** 1 * y_inter_scaled ** 1,
                              y_inter_scaled ** 2,
                              x_inter_scaled ** 3,
                              x_inter_scaled ** 2 * y_inter_scaled ** 1,
                              x_inter_scaled ** 1 * y_inter_scaled ** 2,
                              y_inter_scaled ** 3]
                             ).T
    elif order == 4:
        a_matrix = np.vstack([np.ones(npoints),
                              x_inter_scaled,
                              y_inter_scaled,
                              x_inter_scaled ** 2,
                              x_inter_scaled ** 1 * y_inter_scaled ** 1,
                              y_inter_scaled ** 2,
                              x_inter_scaled ** 3,
                              x_inter_scaled ** 2 * y_inter_scaled ** 1,
                              x_inter_scaled ** 1 * y_inter_scaled ** 2,
                              y_inter_scaled ** 3,
                              x_inter_scaled ** 4,
                              x_inter_scaled ** 3 * y_inter_scaled ** 1,
                              x_inter_scaled ** 2 * y_inter_scaled ** 2,
                              x_inter_scaled ** 1 * y_inter_scaled ** 3,
                              y_inter_scaled ** 4]
                             ).T
    elif order == 5:
        a_matrix = np.vstack([np.ones(npoints),
                              x_inter_scaled,
                              y_inter_scaled,
                              x_inter_scaled ** 2,
                              x_inter_scaled ** 1 * y_inter_scaled ** 1,
                              y_inter_scaled ** 2,
                              x_inter_scaled ** 3,
                              x_inter_scaled ** 2 * y_inter_scaled ** 1,
                              x_inter_scaled ** 1 * y_inter_scaled ** 2,
                              y_inter_scaled ** 3,
                              x_inter_scaled ** 4,
                              x_inter_scaled ** 3 * y_inter_scaled ** 1,
                              x_inter_scaled ** 2 * y_inter_scaled ** 2,
                              x_inter_scaled ** 1 * y_inter_scaled ** 3,
                              y_inter_scaled ** 4,
                              x_inter_scaled ** 5,
                              x_inter_scaled ** 4 * y_inter_scaled ** 1,
                              x_inter_scaled ** 3 * y_inter_scaled ** 2,
                              x_inter_scaled ** 2 * y_inter_scaled ** 3,
                              x_inter_scaled ** 1 * y_inter_scaled ** 4,
                              y_inter_scaled ** 5],
                             ).T
    elif order == 6:
        a_matrix = np.vstack([np.ones(npoints),
                              x_inter_scaled,
                              y_inter_scaled,
                              x_inter_scaled ** 2,
                              x_inter_scaled ** 1 * y_inter_scaled ** 1,
                              y_inter_scaled ** 2,
                              x_inter_scaled ** 3,
                              x_inter_scaled ** 2 * y_inter_scaled ** 1,
                              x_inter_scaled ** 1 * y_inter_scaled ** 2,
                              y_inter_scaled ** 3,
                              x_inter_scaled ** 4,
                              x_inter_scaled ** 3 * y_inter_scaled ** 1,
                              x_inter_scaled ** 2 * y_inter_scaled ** 2,
                              x_inter_scaled ** 1 * y_inter_scaled ** 3,
                              y_inter_scaled ** 4,
                              x_inter_scaled ** 5,
                              x_inter_scaled ** 4 * y_inter_scaled ** 1,
                              x_inter_scaled ** 3 * y_inter_scaled ** 2,
                              x_inter_scaled ** 2 * y_inter_scaled ** 3,
                              x_inter_scaled ** 1 * y_inter_scaled ** 4,
                              y_inter_scaled ** 5,
                              x_inter_scaled ** 6,
                              x_inter_scaled ** 5 * y_inter_scaled ** 1,
                              x_inter_scaled ** 4 * y_inter_scaled ** 2,
                              x_inter_scaled ** 3 * y_inter_scaled ** 3,
                              x_inter_scaled ** 2 * y_inter_scaled ** 4,
                              x_inter_scaled ** 1 * y_inter_scaled ** 5,
                              y_inter_scaled ** 6],
                             ).T
    elif order == 7:
        a_matrix = np.vstack([np.ones(npoints),
                              x_inter_scaled,
                              y_inter_scaled,
                              x_inter_scaled ** 2,
                              x_inter_scaled ** 1 * y_inter_scaled ** 1,
                              y_inter_scaled ** 2,
                              x_inter_scaled ** 3,
                              x_inter_scaled ** 2 * y_inter_scaled ** 1,
                              x_inter_scaled ** 1 * y_inter_scaled ** 2,
                              y_inter_scaled ** 3,
                              x_inter_scaled ** 4,
                              x_inter_scaled ** 3 * y_inter_scaled ** 1,
                              x_inter_scaled ** 2 * y_inter_scaled ** 2,
                              x_inter_scaled ** 1 * y_inter_scaled ** 3,
                              y_inter_scaled ** 4,
                              x_inter_scaled ** 5,
                              x_inter_scaled ** 4 * y_inter_scaled ** 1,
                              x_inter_scaled ** 3 * y_inter_scaled ** 2,
                              x_inter_scaled ** 2 * y_inter_scaled ** 3,
                              x_inter_scaled ** 1 * y_inter_scaled ** 4,
                              y_inter_scaled ** 5,
                              x_inter_scaled ** 6,
                              x_inter_scaled ** 5 * y_inter_scaled ** 1,
                              x_inter_scaled ** 4 * y_inter_scaled ** 2,
                              x_inter_scaled ** 3 * y_inter_scaled ** 3,
                              x_inter_scaled ** 2 * y_inter_scaled ** 4,
                              x_inter_scaled ** 1 * y_inter_scaled ** 5,
                              y_inter_scaled ** 6,
                              x_inter_scaled ** 7,
                              x_inter_scaled ** 6 * y_inter_scaled ** 1,
                              x_inter_scaled ** 5 * y_inter_scaled ** 2,
                              x_inter_scaled ** 4 * y_inter_scaled ** 3,
                              x_inter_scaled ** 3 * y_inter_scaled ** 4,
                              x_inter_scaled ** 2 * y_inter_scaled ** 5,
                              x_inter_scaled ** 1 * y_inter_scaled ** 6,
                              y_inter_scaled ** 7],
                             ).T
    else:
        raise ValueError("Invalid order=" + str(order))
    poltrans = transform.PolynomialTransform(
        np.vstack(
            [np.linalg.lstsq(a_matrix, x_orig_scaled, rcond=None)[0],
             np.linalg.lstsq(a_matrix, y_orig_scaled, rcond=None)[0]]
        )
    )

    # reverse normalization to recover coefficients of the
    # transformation in the correct system
    factor = np.zeros_like(poltrans.params[0])
    k = 0
    for i in range(order + 1):
        for j in range(i + 1):
            factor[k] = (x_scale ** (i - j)) * (y_scale ** j)
            k += 1
    aij = poltrans.params[0] * factor / x_scale
    bij = poltrans.params[1] * factor / y_scale

    # show results
    if abs(debugplot) >= 10:
        print(">>> u=u(x,y) --> aij:\n", aij)
        print(">>> v=v(x,y) --> bij:\n", bij)

    if abs(debugplot) % 10 != 0:
        ax = ximplotxy(x_orig_scaled, y_orig_scaled,
                       show=False,
                       **{'marker': 'o',
                          'label': '(u,v) coordinates', 'linestyle': ''})
        dum = list(zip(x_orig_scaled, y_orig_scaled))
        for idum in range(len(dum)):
            ax.text(dum[idum][0], dum[idum][1], str(idum + 1), fontsize=10,
                    horizontalalignment='center',
                    verticalalignment='bottom', color='black')
        ax.plot(x_inter_scaled, y_inter_scaled, 'o',
                label="(x,y) coordinates")
        dum = list(zip(x_inter_scaled, y_inter_scaled))
        for idum in range(len(dum)):
            ax.text(dum[idum][0], dum[idum][1], str(idum + 1), fontsize=10,
                    horizontalalignment='center',
                    verticalalignment='bottom', color='grey')
        xmin = np.concatenate((x_orig_scaled, x_inter_scaled)).min()
        xmax = np.concatenate((x_orig_scaled, x_inter_scaled)).max()
        ymin = np.concatenate((y_orig_scaled, y_inter_scaled)).min()
        ymax = np.concatenate((y_orig_scaled, y_inter_scaled)).max()
        dx = xmax - xmin
        xmin -= dx / 20
        xmax += dx / 20
        dy = ymax - ymin
        ymin -= dy / 20
        ymax += dy / 20
        ax.set_xlim([xmin, xmax])
        ax.set_ylim([ymin, ymax])
        ax.set_xlabel("pixel (normalized coordinate)")
        ax.set_ylabel("pixel (normalized coordinate)")
        ax.set_title("compute distortion")
        ax.legend()
        pause_debugplot(debugplot, pltshow=True)

    return aij, bij


def fmap(order, aij, bij, x, y):
    """Evaluate the 2D polynomial transformation.

    u = sum[i=0:order]( sum[j=0:i]( a_ij * x**(i - j) * y**j ))
    v = sum[i=0:order]( sum[j=0:i]( b_ij * x**(i - j) * y**j ))

    Parameters
    ----------
    order : int
        Order of the polynomial transformation.
    aij : numpy array
        Polynomial coefficents corresponding to a_ij.
    bij : numpy array
        Polynomial coefficents corresponding to b_ij.
    x : numpy array or float
        X coordinate values where the transformation is computed. Note
        that these values correspond to array indices.
    y : numpy array or float
        Y coordinate values where the transformation is computed. Note
        that these values correspond to array indices.

    Returns
    -------
    u : numpy array or float
        U coordinate values.
    v : numpy array or float
        V coordinate values.

    """

    u = np.zeros_like(x)
    v = np.zeros_like(y)

    k = 0
    for i in range(order + 1):
        for j in range(i + 1):
            u += aij[k] * (x ** (i - j)) * (y ** j)
            v += bij[k] * (x ** (i - j)) * (y ** j)
            k += 1

    return u, v


def coef_fmap_fixed_x(order, aij, bij, x_value):
    """Return coefficients for 1D polynomial, fixing x.

    Parameters
    ----------
    order : int
        Order of the 2D polynomial transformation.
    aij : numpy array
        Polynomial coefficents corresponding to a_ij.
    bij : numpy array
        Polynomial coefficents corresponding to b_ij.
    x_value : float
        Fixed x value.

    Returns
    -------
    coeff_u : numpy array
        Polynomial coefficients of the 1D polynomial corresponding
        to the U coordinate.
    coeff_v : numpy array
        Polynomial coefficients of the 1D polynomial corresponding
        to the U coordinate.

    """

    coeff_u = np.zeros(order + 1)
    coeff_v = np.zeros(order + 1)

    k = 0
    for i in range(order + 1):
        for j in range(i + 1):
            coeff_u[j] += aij[k] * x_value ** (i - j)
            coeff_v[j] += bij[k] * x_value ** (i - j)
            k += 1

    return coeff_u, coeff_v


def coef_fmap_fixed_y(order, aij, bij, y_value):
    """Return coefficients for 1D polynomial, fixing y.

    Parameters
    ----------
    order : int
        Order of the 2D polynomial transformation.
    aij : numpy array
        Polynomial coefficents corresponding to a_ij.
    bij : numpy array
        Polynomial coefficents corresponding to b_ij.
    y_value : float
        Fixed y value.

    Returns
    -------
    coeff_u : numpy array
        Polynomial coefficients of the 1D polynomial corresponding
        to the U coordinate.
    coeff_v : numpy array
        Polynomial coefficients of the 1D polynomial corresponding
        to the U coordinate.

    """

    coeff_u = np.zeros(order + 1)
    coeff_v = np.zeros(order + 1)

    k = 0
    for i in range(order + 1):
        for j in range(i + 1):
            coeff_u[i - j] += aij[k] * y_value ** j
            coeff_v[i - j] += bij[k] * y_value ** j
            k += 1

    return coeff_u, coeff_v


def ncoef_fmap(order):
    """Expected number of coefficients in a 2D transformation of a given order.

    Parameters
    ----------
    order : int
        Order of the 2D polynomial transformation.

    Returns
    -------
    ncoef : int
        Expected number of coefficients.

    """

    ncoef = 0
    for i in range(order + 1):
        for j in range(i + 1):
            ncoef += 1
    return ncoef


def order_fmap(ncoef):
    """Compute order corresponding to a given number of coefficients.

    Parameters
    ----------
    ncoef : int
        Number of coefficients.

    Returns
    -------
    order : int
        Order corresponding to the provided number of coefficients.

    """

    loop = True
    order = 1
    while loop:
        loop = not (ncoef == ncoef_fmap(order))
        if loop:
            order += 1
            if order > NMAX_ORDER:
                print('No. of coefficients: ', ncoef)
                raise ValueError(
                    "order > " + str(NMAX_ORDER) + " not implemented")
    return order


def rectify2d(image2d, aij, bij, resampling,
              naxis1out=None, naxis2out=None,
              ioffx=None, ioffy=None,
              debugplot=0):
    """Rectify image applying the provided 2D transformation.

    The rectified image correspond to the transformation given by:
        u = sum[i=0:order]( sum[j=0:i]( a_ij * x**(i - j) * y**j ))
        v = sum[i=0:order]( sum[j=0:i]( b_ij * x**(i - j) * y**j ))

    Parameters
    ----------
    image2d : 2d numpy array
        Initial image.
    aij : 1d numpy array
        Coefficients a_ij of the transformation.
    bij : 1d numpy array
        Coefficients b_ij of the transformation.
    resampling : int
        1: nearest neighbour, 2: flux preserving interpolation.
    naxis1out : int or None
        X-axis dimension of output image.
    naxis2out : int or None
        Y-axis dimension of output image.
    ioffx : int
        Integer offset in the X direction.
    ioffy : int
        Integer offset in the Y direction.
    debugplot : int
        Determines whether intermediate computations and/or plots
        are displayed. The valid codes are defined in
        numina.array.display.pause_debugplot

    Returns
    -------
    image2d_rect : 2d numpy array
        Rectified image.

    """

    # protections
    ncoef = len(aij)
    if len(bij) != ncoef:
        raise ValueError("aij and bij lengths are different!")

    # order of the polynomial transformation
    order = order_fmap(ncoef)
    if abs(debugplot) >= 10:
        print('--> rectification order:', order)

    # initial image dimension
    naxis2, naxis1 = image2d.shape

    # output image dimension
    if naxis1out is None:
        naxis1out = naxis1
    if naxis2out is None:
        naxis2out = naxis2
    if ioffx is None:
        ioffx = 0
    if ioffy is None:
        ioffy = 0

    if resampling == 1:
        # pixel coordinates (rectified image); since the fmap function
        # below requires floats, these arrays must use dtype=float
        j = np.arange(0, naxis1out, dtype=float) - ioffx
        i = np.arange(0, naxis2out, dtype=float) - ioffy
        # the cartesian product of the previous 1D arrays could be stored
        # as np.transpose([xx,yy]), where xx and yy are computed as follows
        xx = np.tile(j, (len(i),))
        yy = np.repeat(i, len(j))
        # compute pixel coordinates in original (distorted) image
        xxx, yyy = fmap(order, aij, bij, xx, yy)
        # round to the nearest integer and cast to integer; note that the
        # rounding still provides a float, so the casting is required
        ixxx = np.rint(xxx).astype(int)
        iyyy = np.rint(yyy).astype(int)
        # determine pixel coordinates within available image
        lxxx = np.logical_and(ixxx >= 0, ixxx < naxis1)
        lyyy = np.logical_and(iyyy >= 0, iyyy < naxis2)
        lok = np.logical_and(lxxx, lyyy)
        # assign pixel values to rectified image
        ixx = xx.astype(int)[lok]
        iyy = yy.astype(int)[lok]
        ixxx = ixxx[lok]
        iyyy = iyyy[lok]
        # initialize result
        image2d_rect = np.zeros((naxis2out, naxis1out), dtype=float)
        # rectified image
        image2d_rect[iyy + ioffy, ixx + ioffx] = image2d[iyyy, ixxx]
    elif resampling == 2:
        # coordinates (rectified image) of the four corners, sorted in
        # anticlockwise order, of every pixel
        j = np.array(
            [[k - 0.5 - ioffx, k + 0.5 - ioffx,
              k + 0.5 - ioffx, k - 0.5 - ioffx] for k in range(naxis1out)]
        )
        i = np.array(
            [[k - 0.5 - ioffy, k - 0.5 - ioffy,
              k + 0.5 - ioffy, k + 0.5 - ioffy] for k in range(naxis2out)]
        )
        xx = np.reshape(np.tile(j, naxis2out), naxis1out * naxis2out * 4)
        yy = np.concatenate([np.reshape(i, naxis2out * 4)] * naxis1out)
        # compute pixel coordinates in original (distorted) image
        xxx, yyy = fmap(order, aij, bij, xx, yy)
        # indices of pixels in the rectified image
        ixx = np.repeat(np.arange(naxis1out), naxis2out)
        iyy = np.tile(np.arange(naxis2out), (naxis1out,))
        # rectified image (using cython function)
        # Important: the cython code expects the 2D array of float64 numbers
        image2d_rect = _resample(image2d.astype(float), xxx, yyy, ixx, iyy, naxis1out, naxis2out)
    else:
        raise ValueError("Sorry, resampling method must be 1 or 2")

    # return result
    return image2d_rect


def shift_image2d(image2d, xoffset=0.0, yoffset=0.0, resampling=2):
    """Shift image applying arbitray X and Y offsets.

    Parameters
    ----------
    image2d : 2d numpy array
        Initial image.
    xoffset : float
        Offset in the X direction.
    yoffset : float
        Offset in the Y direction.
    resampling : int
        1: nearest neighbour, 2: flux preserving interpolation.

    Returns
    -------
    image2d_shifted : 2d numpy array
        Rectified image.

    """

    aij = np.array([-xoffset, 1.0, 0.0], dtype=float)
    bij = np.array([-yoffset, 0.0, 1.0], dtype=float)
    image2d_shifted = rectify2d(image2d, aij, bij, resampling=resampling)
    return image2d_shifted


def rotate_image2d(image2d, theta_deg, xcenter, ycenter, fscale=1.0, resampling=2):
    """Shift image applying arbitray X and Y offsets.

    Parameters
    ----------
    image2d : 2d numpy array
        Initial image.
    theta_deg : float
        Rotation angle (positive values correspond to counter-clockwise
        angles).
    xcenter : float
        X coordinate of center of rotation, in pixel coordinates (i.e.,
        the image X coordinates run from 0.5 to NAXIS1+0.5).
    ycenter : float
        Y coordinate of center of rotation, in pixel coordinates (i.e.,
        the image Y coordinates run from 0.5 to NAXIS2+0.5).
    fscale : float
        Scale factor (1.0: no change in scale).
    resampling : int
        1: nearest neighbour, 2: flux preserving interpolation.

    Returns
    -------
    image2d_shifted : 2d numpy array
        Rectified image.

    """

    f = 1/fscale
    theta_rad = theta_deg * np.pi/180
    costheta = np.cos(theta_rad)
    sintheta = np.sin(theta_rad)
    xc = xcenter - 1.0
    yc = ycenter - 1.0
    aij = [-f*xc*costheta-f*yc*sintheta+xc, f*costheta, f*sintheta]
    bij = [f*xc*sintheta-f*yc*costheta+yc, -f*sintheta, f*costheta]
    image2d_rotated = rectify2d(image2d, aij, bij, resampling=resampling)
    return image2d_rotated



#
# Copyright 2011-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""DRP system-wide loader"""

import sys

import backports.entry_points_selectable as beps

from .drpbase import DrpGeneric


class DrpSystem(DrpGeneric):
    """Load DRPs from the system."""

    def __init__(self, entry_point='numina.pipeline.1'):
        self.entry = entry_point
        super(DrpSystem, self).__init__()

    def load(self):
        """Load all available DRPs in 'entry_point'."""
        for drpins in self.iload(self.entry):
            self.drps[drpins.name] = drpins

        return self

    @classmethod
    def load_drp(cls, name, entry_point='numina.pipeline.1'):
        """Load all available DRPs in 'entry_point'."""

        for drpins in cls.iload(entry_point):
            if drpins.name == name:
                return drpins
        else:
            raise KeyError(f'{name}')

    @classmethod
    def iload(cls, entry_point='numina.pipeline.1'):
        """Load all available DRPs in 'entry_point'."""

        for entry in beps.entry_points(group=entry_point):
            try:
                drp_loader = entry.load()
                drpins = drp_loader()
                if cls.instrumentdrp_check(drpins, entry.name):
                    yield drpins
            except Exception as error:
                print('Problem loading', entry, file=sys.stderr)
                print("Error is: ", error, file=sys.stderr)





import numpy

import numina.array.combine as c


def no_test_issue34():

    data1 = [[1, 2, 3]]
    data2 = [[8, 2, 3]]
    data3 = [[1, 2, 3]]
    data4 = [[9, 2, 3]]
    result = c.median([data1, data2, data3, data4])
    assert numpy.allclose(result[0, 0, 0], 4.5)



#
# Copyright 2019-2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#


import typing
import warnings

if typing.TYPE_CHECKING:
    from numina.instrument.configorigin import ElementOrigin
    from numina.instrument.hwdevice import DeviceBase

from numina.instrument.hwdevice import HWDevice


class Robot(HWDevice):
    def __init__(
        self,
        name: str,
        model,
        idx: int,
        active=True,
        parent: "DeviceBase | None" = None,
    ):
        super().__init__(name, parent=parent)
        self.idx = idx
        self.angle_ = 0
        self.active_ = active
        self.model = model

    def configure_me_with_header(self, hdr):
        with self.state.managed_configure():
            self.angle = hdr.get(f"R{self.idx:02d}_ANGL", 0)
            self.active = hdr.get(f"R{self.idx:02d}_ACTV", True)

    @property
    def angle(self):
        return self.angle_

    @angle.setter
    def angle(self, value: float) -> None:
        if self.active_ or self.state.is_configuring:
            self.angle_ = value
        else:
            # warning, Robot is inactive
            warnings.warn(f"Robot {self.idx} is inactive")
            pass

    @property
    def active(self):
        return self.active_

    @active.setter
    def active(self, value):
        self.active_ = value


class RobotPos(HWDevice):
    def __init__(
        self,
        name: str,
        robot_model,
        origin: "ElementOrigin | None" = None,
        parent: "DeviceBase | None" = None,
    ):
        super().__init__(name, origin=origin, parent=parent)
        self.robot_model = robot_model
        active = True
        for idx, model in enumerate(robot_model["ypos"]):
            robo = Robot(f"arm_{idx}", model, idx, active=active)
            robo.set_parent(self)

    @classmethod
    def from_component(
        cls,
        name: str,
        comp_id: str,
        origin: "ElementOrigin | None" = None,
        parent: "DeviceBase | None" = None,
        properties=None,
        setup=None,
    ) -> "RobotPos":

        robot_model = setup.values["robotmodel"]
        obj = cls.__new__(cls)
        obj.__init__(comp_id, robot_model, origin, parent)
        return obj
